# cython: language_level=3
cdef extern from "eob_parameters.h":
    int PN_limit
    int ell_max



cdef class PhysicalParams:
    cdef public double m_1,m_2,nu,M,X_1,X_2,delta,chi_1,chi_2,chi_S,chi_A,a,ap,am
    cdef public double chi1_L, chi2_L,a1,a2, H_val
    cdef public double[:] chi1_v
    cdef public double[:] chi2_v
    cdef public double[:] lN
    cdef public double omega,omega_circ #,omega_start
    cpdef update_spins(self,double chi_1,double chi_2)
    cdef _compute_derived_quants(self)

'''
cdef class CalibCoeffs:
    cdef public double a6,d5,dSO,dSS
    cdef public double flagNLOSO,flagNLOSO2,flagNLOSO3,flagNLOSS,flagNLOSS2,flagS3
'''
cdef class CalibCoeffs:
    cdef public dict dc

cdef class Dynamics:
    cdef public double[:] p_circ

cdef class FluxParams:
    # Coefficients entering the amplitude residual
    cdef public double[:,:,:] rho_coeffs
    cdef public double[:,:,:] rho_coeffs_log
    cdef public double[:,:,:] f_coeffs
    # Notice that f_coeffs_vh are allowed to be complex
    cdef public double complex[:,:,:] f_coeffs_vh

    # Coefficients entering the phase residual
    cdef public double complex[:,:,:] delta_coeffs
    cdef public double complex[:,:,:] delta_coeffs_vh

    # Newtonian prefixes
    # For waveform (complex)
    cdef public double complex[:,:] prefixes
    # For flux (real)
    cdef public double[:,:] prefixes_abs

    # These will hold the *results* of computing various pieces
    # of the waveform
    cdef public double[:,:] Tlm
    cdef public double complex[:,:] rholm
    cdef public double complex[:,:] deltalm

    # Holds nqc coeffs *for the flux*
    cdef public double[:,:,:] nqc_coeffs

    # Holds any extra calibration params that will be used in rholm
    # CAVEAT EMPTOR: this does *exactly* what you tell it to.
    cdef public double [:,:,:] extra_coeffs
    cdef public double [:,:,:] extra_coeffs_log

    cdef public list special_modes
    cdef public bint rho_initialized

    # Include more PN information than in SEOBNRv4HM
    cdef public bint extra_PN_terms

cdef class EOBParams:
    cdef public PhysicalParams p_params
    cdef public CalibCoeffs c_coeffs
    cdef public FluxParams flux_params
    cdef public Dynamics dynamics
    cdef public list mode_array
    cdef public bint aligned



#!/usr/bin/env python
"""This file contains the default settings for various models
Each default setting is a *function* the resturns a dictionary.
To be used by models in models.py
"""

from typing import Any, Dict

import lal


def default_NR_LVC_settings() -> Dict[Any, Any]:
    """Return the default settings for the LVC NR approximant

    Returns:
        Dict[Any]: settings dictionary
    """
    settings = dict(
        M=50.0,  # Total mass in solar masses
        distance=400 * lal.PC_SI * 1e6,  # Distance in meters
        dt=2.4627455127717882e-05,  # Sampling rate, in SI units
        modes=[(2, 2)],  # Active modes
    )
    return settings


def default_NR_SXS_settings() -> Dict[Any, Any]:
    """Return the default settings for SXS NR waveforms

    Returns:
        Dict[Any]: settings dictionary
    """
    settings = dict(
        modes=[(2, 2)],  # Active modes
        extrap_order=2,  # Extrapolation order
        alignment_interval_start=2,  # Which peak in Re(h22) is the start
        alignment_interval_end=12,  # Which peak in Re(h22) is the end
        dt=0.1,
    )
    return settings


def default_NR_RIT_settings() -> Dict[Any, Any]:
    """Return the default settings for SXS NR waveforms

    Returns:
        Dict[Any]: settings dictionary
    """
    settings = dict(
        modes=[(2, 2)],  # Active modes
        alignment_interval_start=2,  # Which peak in Re(h22) is the start
        alignment_interval_end=6,  # Which peak in Re(h22) is the end
    )
    return settings


def default_SEOBNRv4HM_settings() -> Dict[Any, Any]:
    """Return the default settings for the SEOBNRv4 approximant

    Returns:
        Dict[Any]: settings dictionary
    """
    settings = dict(
        M=50.0,  # Total mass in solar masses
        distance=400 * lal.PC_SI * 1e6,  # Distance in meters
        dt=2.4627455127717882e-05,  # Sampling rate, in SI units
        debug=False,  # Run in debug mode
    )
    return settings


def default_NRHybSur3dq8_settings() -> Dict[Any, Any]:
    """Return the default settings for the NRHybSur3dq8 approximant

    Returns:
        Dict[Any]: settings dictionary
    """
    settings = dict(
        M=50.0,  # Total mass in solar masses
        dt=0.1,  # Sampling rate in geometric units
        alignment_interval_start=2,  # Which peak in Re(h22) is the start
        alignment_interval_end=12,  # Which peak in Re(h22) is the end
    )
    return settings


def default_NRHybSur2dq15_settings() -> Dict[Any, Any]:
    """Return the default settings for the NRHybSur2dq15 approximant

    Returns:
        Dict[Any]: settings dictionary
    """
    settings = dict(
        M=50.0,  # Total mass in solar masses
        dt=0.1,  # Sampling rate in geometric units
        alignment_interval_start=2,  # Which peak in Re(h22) is the start
        alignment_interval_end=12,  # Which peak in Re(h22) is the end
    )
    return settings


def default_unfaithfulness_phys_v4_settings() -> Dict[Any, Any]:
    settings = dict(
        sigma=0.001,  # the sigma to use in likelihood
    )
    return settings


def default_amplitude_fractional_difference_settings() -> Dict[Any, Any]:
    settings = dict(
        fraction=0.1,  # fraction of the waveform starting at peak going back to consider
        sigma=0.005,  # the sigma to use in likelihood
    )
    return settings


def default_unfaithfulness_flat_settings() -> Dict[Any, Any]:
    settings = dict(sigma=0.01, leading_order=True)
    return settings


def default_delta_time_to_merger_settings() -> Dict[Any, Any]:
    settings = dict(sigma=5)
    return settings


def default_metric_list() -> Dict[Any, Any]:
    return []
    # return [' UnfaithfulnessPhysV4','DeltaTimeToMerger']



from pyseobnr.eob.hamiltonian.Hamiltonian_C cimport Hamiltonian_C

cpdef fin_diff_derivative(
    x,
    y,
    n=?,
)

cpdef cumulative_integral(
    x: np.array,
    y: np.array,
    order = ?,
)

cpdef Kerr_ISCO(
    double chi1,
    double chi2,
    double m1,
    double m2,
)

cpdef compute_adiabatic_solution(
    double[:] r,
    Hamiltonian_C H,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    double[::1] q,
    double[::1] p,
    double tol=?,
)


import argparse

import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path
import importlib
import os,sys

from pathos.multiprocessing import ProcessingPool as Pool

from pyseobnr.generate_waveform import generate_modes_opt
from pyseobnr.auxiliary.sanity_checks.parameters import parameters_random_fast
from pyseobnr.auxiliary.sanity_checks.single_waveform_tests import amplitude_hierarchy_test



def amplitude_hierarchy(q: float, chi1x: float,  chi1y: float,  chi1z: float,  chi2x: float,  chi2y: float,  chi2z: float,  plots: bool = True):

    """
    Generate a SEOBNRv5 waveform with parameters (q, chi1, chi2) and run
    amplitude_hierarchy_test to check that the amplitude of the subdominant
    modes is less than the amplitude of the (2,2) mode until the peak of the (2,2) modes

    Return (0,0,0) if the test passed, (q, chi1, chi2) otherwise
    """

    omega0 = 0.015
    omega_start = omega0
    mt = 60.
    delta_t = 1./8192.

    chi_1 = [chi1x, chi1y, chi1z]
    chi_2 = [chi2x, chi2y, chi2z]


    try:

        ell_max = 5
        settings = {'ell_max':ell_max,'beta_approx':None,'M':mt,"dt":delta_t,"return_coprec":True,
        "postadiabatic": True,
        "postadiabatic_type": "analytic",
        "initial_conditions":"adiabatic",
        "initial_conditions_postadiabatic_type":"analytic"}
        _, _, model = generate_modes_opt(q,chi_1,chi_2,omega0,approximant='SEOBNRv5PHM',
                                   debug=True,settings=settings)

        t_h = model.t
        hlms_cop = model.coprecessing_modes

        # Specify the modes so that the negative m modes are not included
        modes_to_test = ["2,2","2,1","3,3","3,2","4,4","4,3","5,5"]
        hlms_dict ={}
        for lm in modes_to_test:
            hlms_dict[lm] = hlms_cop[lm]

        habs22 = np.abs(hlms_cop["2,2"])
        t_merger = t_h[np.argmax(habs22)]

        try:
            amplitude_hierarchy_test(t_h, hlms_dict)
            return (0, 0, 0, 0, 0, 0, 0)
        except AssertionError:
            print(
                f"Amplitude of one subdominant mode is higher than the (2,2) mode the following parameters: q = {q}, chi1x = {chi1x}, chi1y = {chi1y}, chi1z = {chi1z}, chi2x = {chi2x}, chi2y = {chi2y}, chi2z = {chi2z}"
            )
            if plots:
                label = f"q{np.round(q,2)}__s{np.round(chi1x,3)}_{np.round(chi1y,3)}_{np.round(chi1z,3)}__s{np.round(chi2x,3)}_{np.round(chi2y,3)}_{np.round(chi2z,3)}"
                plt_dir = "./plots_hierarchy"
                Path(plt_dir).mkdir(parents=True, exist_ok=True)

                for key in model.modes_list:
                    plt.plot(
                        t_h, np.abs(hlms_dict[f"{key[0]},{key[1]}"]), label=f"[{key}]"
                    )

                plt.xlabel("t/M")
                plt.legend(loc="best")
                plt.axvline(t_merger, ls="--")
                plt.xlim(t_merger - 30, t_merger + 30)
                plt.xlabel("$t/M$")
                plt.ylabel("$|h_{\ell m}|$")
                plt.title(f"{label}")
                plt.savefig(f"{plt_dir}/check_amplitude_hierarchy_{label}.png", dpi=150)
                plt.close()
            return (q, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z)
    except IndexError:
        print(
            f"Error for the following parameters: q = {q}, chi1x = {chi1x}, chi1y = {chi1y}, chi1z = {chi1z}, chi2x = {chi2x}, chi2y = {chi2y}, chi2z = {chi2z}"
        )


def process_one_case_prec(input):
    q, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z, plots = input

    (
        q_bad,
        chi1x_bad,
        chi1y_bad,
        chi1z_bad,
        chi2x_bad,
        chi2y_bad,
        chi2z_bad
    ) = amplitude_hierarchy(q, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z, plots)
    return np.array([q_bad, chi1x_bad, chi1y_bad, chi1z_bad, chi2x_bad, chi2y_bad, chi2z_bad])

def parameters_random_fast_prec(
    seed: int,
    N: int,
    qmin: float,
    qmax: float,
    a1min: float,
    a1max: float,
    a2min: float,
    a2max: float,
):
    """
    Generate random parameters for q, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z

    Parameters
    ----------
    N:
        Number of parameters to generate
    qmin, qmax:
        Bounds on the q parameter
    a1min, a1max, a2min, a2max:
        Bounds on the a parameter

    Return parameters of the waveforms
    """


    np.random.seed(seed)
    q = np.random.uniform(qmin, qmax, N)

    np.random.seed(seed+1)
    a1 = np.random.uniform(a1min, a1max, N)
    np.random.seed(seed+2)
    a2 = np.random.uniform(a2min, a2max, N)


    np.random.seed(seed+3)
    theta1 = np.random.uniform(0,np.pi,N)
    np.random.seed(seed+4)
    theta2 = np.random.uniform(0,np.pi,N)

    np.random.seed(seed+5)
    phi1 = np.random.uniform(0,2*np.pi,N)
    np.random.seed(seed+6)
    phi2 = np.random.uniform(0,2*np.pi,N)

    chi1x = a1*np.sin(theta1)*np.cos(phi1)
    chi1y = a1*np.sin(theta1)*np.sin(phi1)
    chi1z = a1*np.cos(theta1)

    chi2x = a2*np.sin(theta2)*np.cos(phi2)
    chi2y = a2*np.sin(theta2)*np.sin(phi2)
    chi2z = a2*np.cos(theta2)

    return (
        np.array(q),
        np.array(chi1x),
        np.array(chi1y),
        np.array(chi1z),
        np.array(chi2x),
        np.array(chi2y),
        np.array(chi2z),
    )


if __name__ == "__main__":

    p = argparse.ArgumentParser(
        description="Check amplitude hierarchy across parameter space"
    )
    p.add_argument("--points", type=int, help="Number of points", default="5000")
    p.add_argument("--q-max", type=float, help="Upper limit on q", default="100.0")
    p.add_argument(
        "--a-max", type=float, help="Upper limit on |chi|", default="0.995"
    )
    p.add_argument(
        "--name",
        type=str,
        help="Name of the output file",
        default="amplitude_hierarchy.dat",
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )
    p.add_argument(
        "--wrapper-path", type=str, help="The path to the wrapper, including name"
    )
    p.add_argument("--n-cpu", type=int, help="Number of cores to use", default=64)
    p.add_argument("--seed", type=int, help="Seed for random generation use", default=150914)
    args = p.parse_args()

    qarr, chi1x_arr, chi1y_arr, chi1z_arr, chi2x_arr, chi2y_arr, chi2z_arr = parameters_random_fast_prec(
        args.seed,
        args.points,
        1.0,
        args.q_max,
        0.,
        args.a_max,
        0.,
        args.a_max,
    )

    lst = [(a, b, c, d, e, f, g, args.plots) for a, b, c, d, e, f, g in zip(qarr, chi1x_arr, chi1y_arr, chi1z_arr, chi2x_arr, chi2y_arr, chi2z_arr)]
    module_name = os.path.basename(args.wrapper_path).split(".")[0]
    # The following hacks things so that the wrapper can be loaded from the specified file
    file_path = args.wrapper_path
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    globals().update(
        {n: getattr(module, n) for n in module.__all__}
        if hasattr(module, "__all__")
        else {k: v for (k, v) in module.__dict__.items() if not k.startswith("_")}
    )

    pool = Pool(args.n_cpu)

    all_means = pool.map(process_one_case_prec, lst)
    all_means = np.array(all_means)

    np.savetxt(args.name, all_means)

    ### Plots ###

    if args.plots:

        plt_dir = "./plots_hierarchy"
        Path(plt_dir).mkdir(parents=True, exist_ok=True)

        res_path = args.name
        res = np.loadtxt(res_path)

        q = res[:, 0]
        idx = np.where(q > 0.0)[0]

        q = q[idx]
        chi1_x = res[:, 1][idx]
        chi1_y = res[:, 2][idx]
        chi1_z = res[:, 3][idx]

        chi2_x = res[:, 4][idx]
        chi2_y = res[:, 5][idx]
        chi2_z = res[:, 6][idx]


        nu = q / (1. + q) ** 2
        m1 = q / (1. + q)
        m2 = 1. / (1. + q)
        ap = m1 * chi1_z + m2 * chi2_z
        am = m1 * chi1_z - m2 * chi2_z

        if len(q) > 0:
            # Cases where modes amplitude hierarchy is not satisfied across parameter space - (q, chieff)
            plt.scatter(q, ap)
            plt.ylabel("$\chi_{\mathrm{eff}}$")
            plt.xlabel("$q$")

            plt.title("Cases where modes amplitude_hierarchy is not satisfied")
            plt.savefig(
                f"{plt_dir}/amplitude_hierarchy.png", bbox_inches="tight", dpi=300
            )
            plt.close()
        else:
            print("Test passed for all cases")



#!/usr/bin/env python3
"""
Computes the aligned-spin initial conditions in polar coordinates. 
"""

import logging
import numpy as np
from scipy.optimize import root, root_scalar

from rich.logging import RichHandler
from rich.traceback import install

# Setup the logger to work with rich
logger = logging.getLogger(__name__)
logger.addHandler(RichHandler(rich_tracebacks=True, markup=True))
logger.setLevel("INFO")
# Setup rich to get nice tracebacks
install()


def IC_cons(u, omega, H, chi_1, chi_2, m_1, m_2):
    """The equations defining the 'conservative'
    part of the QC initial conditions, namely
    for r and pphi.

    This is Eq(60) in
    https://journals.aps.org/prd/pdf/10.1103/PhysRevD.104.024046

    Args:
        u ([np.ndarray]): The unknowns, r,pphi
        omega ([float): Desired starting orbital frequency, in gemoetric units
        H (function): The Hamiltonian to use (an instance of Hamiltonian class)
        chi_1 (float): z-component of the primary spin
        chi_2 (float): z-component of the secondary spin
        m_1 (float): mass of the primary
        m_2 (float): mass of the secondary

    Returns:
        [np.ndarray]: The desired equations evaluated at u
    """
    r, pphi = u

    q = np.array([r, 0.0])
    p = np.array([0.0, pphi])
    grad = H.dynamics(q, p, chi_1, chi_2, m_1, m_2)
    dHdr = grad[0]
    dHdpphi = grad[3]
    diff = np.array([dHdpphi - omega, dHdr])
    return diff


def IC_diss(u, r, pphi, H, RR, chi_1, chi_2, m_1, m_2, params):
    """Initial conditions for the "dissipative" part,
    namely pr.

    This is Eq(68) in
    https://journals.aps.org/prd/pdf/10.1103/PhysRevD.104.024046

    Note that RR_f[1] is 1/Omega*dE/dt

    Args:
        u (float): Guess for pr
        r (floart): Starting separation
        pphi (float): Starting angular momentum
        H (Hamiltonian): The Hamiltonian object to use
        RR (function): Function that returns the RR force. Must have same signature as the Hamiltonian
        chi_1 (float): z-component of the dimensionless spin of primary
        chi_2 (float): z-component of the dimensionless spin of secondary
        m_1 (float): Mass of primary
        m_2 (float): Mass of secondary

    Returns:
        float: The equation for pr
    """
    pr = u
    q = np.array([r, 0.0])
    p = np.array([pr, pphi])
    hess = H.hessian(q, p, chi_1, chi_2, m_1, m_2)
    d2Hdr2 = hess[0, 0]
    d2HdrdL = hess[3, 0]
    dLdr = -d2Hdr2 / d2HdrdL
    p_circ = np.array([0.0, p[1]])
    dynamics = H.dynamics(q, p, chi_1, chi_2, m_1, m_2)
    H_val = dynamics[4]
    omega = dynamics[3]
    omega_circ = H.omega(q, p_circ, chi_1, chi_2, m_1, m_2)
    RR_f = RR.RR(q, p, omega, omega_circ, H_val, params)
    xi = dynamics[5]

    rdot = 1 / xi * RR_f[1] / dLdr
    dHdpr = dynamics[2]
    return rdot - dHdpr


def computeIC_opt(omega, H, RR, chi_1, chi_2, m_1, m_2, **kwargs):
    """Compute the initial conditions for an aligned-spin BBH binary

    Args:
        omega (float): Initial *orbital* frequency in geometric units
        H (Hamiltonian): The Hamiltonian object to use
        RR (function): The RR force to use
        chi_1 (float): z-component of the dimensionless spin of primary
        chi_2 (float): z-component of the dimensionless spin of secondary
        m_1 (float): Mass of primary
        m_2 (float): Mass of secondary

    Returns:
        tuple: The initial conditions: (r,pphi,pr)
    """
    # Initial guess from Newtonian gravity
    r_guess = omega ** (-2.0 / 3)
    z = [r_guess, np.sqrt(r_guess)]
    # print(f"Initial guess is {z}")
    # The conservative bit: solve for r and pphi
    res_cons = root(IC_cons, z, args=(omega, H, chi_1, chi_2, m_1, m_2), tol=6e-12)
    if not res_cons.success:
        logger.error(
            f"The solution fo the conservative part of intial conditions failed for"
            f" m1={m_1},m2={m_2},chi1={chi_1},chi2={chi_2},omega={omega}"
        )

    r0, pphi0 = res_cons.x
    # print(f"Computed conservative stuff, {r0},{pphi0}")

    res_diss = root_scalar(
        IC_diss,
        bracket=[-3e-2, 0],
        args=(r0, pphi0, H, RR, chi_1, chi_2, m_1, m_2, kwargs["params"]),
        xtol=1e-12,
        rtol=1e-10,
    )
    if not res_diss.converged:
        logger.error(
            "The solution fo the dissipative part of intial conditions failed for"
            " m1={m1},m2={m2},chi1={chi_1},chi2={chi_2},omega={omega}"
        )
    # Now do the dissipative bit: solve for pr
    pr0 = res_diss.root
    return r0, pphi0, pr0



from . import eob
from .eob.dynamics import *
from .eob.hamiltonian import *
from .eob.fits import *
from . import generate_waveform


try:
    from ._version import version as __version__
except ModuleNotFoundError:
    __version__ = ""



cimport cython
from pyseobnr.eob.utils.containers cimport EOBParams

cdef class Hamiltonian_C:
    def __cinit__(self,EOBParams):
        pass


    cpdef _call(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2):
        pass
    cpdef grad(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2):
        pass
    cpdef hessian(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2):
        pass
    cpdef double csi(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2):
        pass
    cpdef  dynamics(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2):
        pass
    cpdef double omega(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2):
        pass
    cpdef auxderivs(self, double[:]q,double[:]p,double chi_1,double chi_2,double m_1,double m_2):
        pass


#!/usr/bin/env python3
"""
Computes the quasi-precessing initial conditions in polar coordinates. 
These have the same equations as the aligned-spin, however:
- the Hamiltonian is different because it depends on several spin degrees of freedom
- appropriate spin projections (and spins themselves) are passed to the Hamiltonian and the RR forces
"""

from typing import Callable
from ..hamiltonian import Hamiltonian
import numpy as np
from scipy.optimize import root, root_scalar
from ..fits.fits_Hamiltonian import dSO as dSO_poly_fit

def IC_cons_augm(u:np.ndarray,
                 omega:float,
                 H: Hamiltonian,
                 chi1_v:np.ndarray,
                 chi2_v:np.ndarray,
                 chi1_LN:float,
                 chi2_LN:float,
                 chi1_L:float,
                 chi2_L:float,
                 m_1:float,
                 m_2:float
                 ):
    """ The equations defining the 'conservative'
    part of the QC initial conditions, namely
    for r and pphi.

    These are Eq(4.8,4.9) in https://arxiv.org/pdf/gr-qc/0508067.pdf

    Args:
        u ([np.ndarray]): The unknowns, r,pphi
        omega ([float): Desired starting orbital frequency, in gemoetric units
        H (function): The Hamiltonian to use (an instance of Hamiltonian class)
        chi1_v (np.ndarray): Dimensionless spin vector of the primary
        chi2_v (np.ndarray): Dimensionless spin vector of the secondary
        chi1_LN (float): Projection of the dimensionless spin vector of the primary onto LN
        chi2_LN (float): Projection of the dimensionless spin vector of the secondary onto LN
        chi1_L (float): Projection of the dimensionless spin vector of the primary onto L
        chi2_L (float): Projection of the dimensionless spin vector of the secondary onto L
        m_1 (float): mass of the primary
        m_2 (float): mass of the secondary

    Returns:
        [np.ndarray]: The desired equations evaluated at u
    """
    r, pphi = u

    q = np.array([r, 0.0])
    p = np.array([0.0, pphi])

    grad = H.dynamics(q, p, chi1_v, chi2_v, m_1, m_2, chi1_LN, chi2_LN, chi1_L, chi2_L)
    dHdr = grad[0]
    dHdpphi = grad[3]
    diff = np.array([dHdpphi - omega, dHdr])
    return diff


def IC_diss_augm(u: np.ndarray,
                 r: float,
                 pphi: float,
                 H: Hamiltonian,
                 RR: Callable,
                 chi1_v: np.ndarray,
                 chi2_v: np.ndarray,
                 chi1_LN: float,
                 chi2_LN: float,
                 chi1_L: float,
                 chi2_L: float,
                 m_1: float,
                 m_2: float,
                 params):
    """Initial conditions for the "dissipative" part,
    namely pr

    This is basically Eq(4.15) in https://arxiv.org/pdf/gr-qc/0508067.pdf
    Args:
        u (float): Guess for pr
        r (floart): Starting separation
        pphi (float): Starting angular momentum
        H (Hamiltonian): The Hamiltonian object to use
        RR (Callable): Function that returns the RR force. Must have same signature as the Hamiltonian
        chi1_v (np.ndarray): Dimensionless spin vector of the primary
        chi2_v (np.ndarray): Dimensionless spin vector of the secondary
        chi1_LN (float): Projection of the dimensionless spin vector of the primary onto LN
        chi2_LN (float): Projection of the dimensionless spin vector of the secondary onto LN
        chi1_L (float): Projection of the dimensionless spin vector of the primary onto L
        chi2_L (float): Projection of the dimensionless spin vector of the secondary onto L
        m_1 (float): mass of the primary
        m_2 (float): mass of the secondary
        params (EOBParams): container with additional useful quantities

    Returns:
        float: The equation for pr
    """
    pr = u
    q = np.array([r, 0.0])
    p = np.array([pr, pphi])
    hess = H.hessian(q, p,chi1_v, chi2_v, m_1, m_2, chi1_LN, chi2_LN, chi1_L, chi2_L)
    d2Hdr2 = hess[0, 0]
    d2HdrdL = hess[3, 0]
    dLdr = -d2Hdr2 / d2HdrdL
    p_circ = np.array([0.0, p[1]])
    dynamics = H.dynamics(q, p, chi1_v, chi2_v, m_1, m_2, chi1_LN, chi2_LN, chi1_L, chi2_L)
    H_val = dynamics[4]
    omega = dynamics[3]
    omega_circ = H.omega(q, p_circ, chi1_v, chi2_v, m_1, m_2, chi1_LN, chi2_LN, chi1_L, chi2_L)
    RR_f = RR.RR(q, p, omega, omega_circ, H_val, params)
    csi = dynamics[5]


    rdot = 1 / csi * RR_f[1] / dLdr
    dHdpr = dynamics[2]
    return rdot - dHdpr

def computeIC_augm(omega: float,
                   H: Hamiltonian,
                   RR: Callable,
                   chi1_v: np.ndarray,
                   chi2_v: np.ndarray,
                   m_1: float,
                   m_2: float,
                   **kwargs):
    """Compute the initial conditions for an aligned-spin BBH binary

    Args:
        omega (float): Initial *orbital* frequency in geometric units
        H (Hamiltonian): The Hamiltonian object to use
        RR (function): The RR force to use
        chi1_v (np.ndarray): Dimensionless spin vector of the primary
        chi2_v (np.ndarray): Dimensionless spin vector of the secondary
        m_1 (float): Mass of primary
        m_2 (float): Mass of secondary

    Returns:
        tuple: The initial conditions: (r,pphi,pr)
    """

    params = kwargs["params"]

    chi1_LN, chi2_LN = params.p_params.chi_1, params.p_params.chi_2
    chi1_L, chi2_L = params.p_params.chi1_L, params.p_params.chi2_L

    # Initial guess from Newtonian gravity
    r_guess = omega ** (-2.0 / 3)
    z = [r_guess, np.sqrt(r_guess)]

    # The conservative bit: solve for r and pphi
    res_cons = root(IC_cons_augm, z, args=(omega, H, chi1_v, chi2_v, chi1_LN, chi2_LN, chi1_L, chi2_L, m_1, m_2), tol=1e-12)
    r0, pphi0 = res_cons.x

    res_diss = root_scalar(
        IC_diss_augm,
        bracket=[-3e-2, 0],
        args=(r0, pphi0, H, RR, chi1_v, chi2_v, chi1_LN, chi2_LN, chi1_L, chi2_L, m_1, m_2,kwargs["params"]),
        xtol=1e-12,
        rtol=1e-10,
    )

    # Now do the dissipative bit: solve for pr
    pr0 = res_diss.root


    # Evaluate omega_circ and the Hamiltonian at the intial conditions
    q = np.array([r0, 0.0])
    p = np.array([pr0, pphi0])
    p_circ = np.array([0.0, p[1]])
    dynamics = H.dynamics(q, p, chi1_v, chi2_v, m_1, m_2, chi1_LN, chi2_LN, chi1_L, chi2_L)
    H_val = dynamics[4]
    omega_circ = H.omega(q, p_circ, chi1_v, chi2_v, m_1, m_2, chi1_LN, chi2_LN, chi1_L, chi2_L)
    params.p_params.omega_circ = omega_circ
    params.p_params.H_val = H_val

    #print(f"Ad. IC : r0 = {r0}, pphi0 = {pphi0}, pr0 = {pr0}")

    return r0, pphi0, pr0



import argparse
import importlib
import os,sys
from mimetypes import MimeTypes
import lal, lalsimulation as lalsim
import numpy as np


import matplotlib.pyplot as plt
from pathlib import Path
from pathos.multiprocessing import ProcessingPool as Pool

from pyseobnr.generate_waveform import generate_modes_opt
from pyseobnr.auxiliary.sanity_checks.parameters import parameters_random_fast
from pyseobnr.auxiliary.sanity_checks.single_waveform_tests import amplitude_hierarchy_test

from pycbc.psd.analytical import aLIGOZeroDetHighPower, aLIGOZeroDetHighPowerGWINC
from pycbc.types import TimeSeries
from pycbc.filter import make_frequency_series
from pycbc.filter.matchedfilter import optimized_match
from pycbc.waveform.utils import taper_timeseries


from typing import Dict, Union, Tuple
def combine_modes(
    iota: float, phi: float, modes_dict: Dict
) -> Tuple[np.array, np.array]:
    """Combine modes to compute the waveform polarizations in the direction
    (iota,np.pi/2-phi)

    Args:
        iota (float): Inclination angle (rad)
        phi (float): Azimuthal angle(rad)
        modes_dict (Dict): Dictionary containing the modes, either time of frequency-domain

    Returns:
        np.array: Waveform in the given direction
    """
    sm = 0.0
    for key in modes_dict.keys():
        #print(key)
        ell, m = [int(x) for x in key.split(",")]
        Ylm0 = lal.SpinWeightedSphericalHarmonic(iota, np.pi / 2 - phi, -2, ell, m)
        sm += Ylm0 * modes_dict[key]

    return np.real(sm), -np.imag(sm)


def timeMtoSec(timeM, mt:float):
    return timeM*mt*lal.MTSUN_SI


def ampNRtoPhysicalTD(ampNR, mt:float,distance:float):
    return ampNR*(lal.C_SI*mt*lal.MTSUN_SI)/distance



def generate_v5PHM_waveform(m1:float, m2:float,
                            s1x:float, s1y:float, s1z:float,
                            s2x:float, s2y:float, s2z:float,
                            f_min:float,
                            iota:float, phi:float, distance:float,
                            delta_t:float, approx:str,
                            ell_max: int = 5):

    mtotal = m1+m2
    q = m1/m2
    omega0 = f_min * (np.pi * (m1 + m2) * lal.MTSUN_SI)

    chi_1 = [s1x, s1y, s1z]
    chi_2 = [s2x, s2y, s2z]

    settings = {'ell_max':ell_max,'beta_approx':None,'M':mtotal,"dt":delta_t,
    "postadiabatic": True,
    "postadiabatic_type": "analytic",
    "initial_conditions":"adiabatic",
    "initial_conditions_postadiabatic_type":"analytic"}
    if approx == 'SEOBNRv5PHM':
        #time, modes, _ = v5P_wrapper.get_EOB_modes(p, ell_max=ell_max)
        time, modes = generate_modes_opt(q,chi_1,chi_2,omega0,approximant=approx,
                           debug=False,settings=settings)

        #amp22 = abs(modes['2,2'])
        #idx_max = np.argmax(amp22)
        #tmax22 = time[idx_max]

        #time -= tmax22

        modes_dict = {}
        for ell in range(2, ell_max + 1):
            for m in range(-ell, ell + 1):
                modes_dict["{},{}".format(ell, m)] = modes["{},{}".format(ell, m)]

    elif approx == 'SEOBNRv5HM':

        time, modes = generate_modes_opt(q,chi_1[-1],chi_2[-1],omega0,approximant=approx,
                                debug=False,settings=settings)

        modes_dict = {}
        for key in modes.keys():
            lm_tag = key.split(',')
            ell,m = lm_tag
            ll = float(ell)
            modes_dict["{},{}".format(ell, m)] = modes[key] # Minus sign to have same convention as v5PHM
            modes_dict["{},-{}".format(ell,m)] = pow(-1.,ll) * np.conjugate(modes[key]) # Minus sign to have same convention as v5PHM

    elif 'IMRPhenomT' in approx:

        time = {}
        modes_dict = {}
        wf_param = lal.CreateDict()
        hp, hc = lalsim.SimInspiralChooseTDWaveform(
                        m1 * lal.MSUN_SI,
                        m2 * lal.MSUN_SI,
                        s1x,
                        s1y,
                        s1z,
                        s2x,
                        s2y,
                        s2z,
                        distance,
                        iota,
                        phi,
                        0.0,
                        0.,
                        0.,
                        delta_t,
                        f_min,
                        f_min,
                        wf_param,
                        lalsim.GetApproximantFromString(approx),
                    )
    else:
        raise NotImplementedError

    if 'SEOBNRv5' in approx:
        hp_NR, hc_NR = combine_modes(iota, phi, modes_dict)

        t_s = timeMtoSec(time, mtotal)
        hp = ampNRtoPhysicalTD(hp_NR,mtotal,distance)
        hc = ampNRtoPhysicalTD(hc_NR,mtotal,distance)

        # Taper
        hp_td = TimeSeries(hp, delta_t=delta_t)
        hc_td = TimeSeries(hc, delta_t=delta_t)
    else:

        # Taper
        hp_td = TimeSeries(hp.data.data, delta_t=delta_t)
        hc_td = TimeSeries(hc.data.data, delta_t=delta_t)

    hp_td = taper_timeseries(hp_td, tapermethod="startend")
    hc_td = taper_timeseries(hc_td, tapermethod="startend")

    return time, modes_dict, hp_td, hc_td

# Reproducible
seed = 150914

mode_list = [(2, 2), (2, 1), (3, 3), (3, 2), (4, 4), (4, 3), (5, 5)]

def mismatch_v5P_strain(
    m1: float,
    m2: float,
    chi1: float,
    chi2: float,
    iota_s: float,
):

    Mt = m1+m2
    q=m1/m2

    if q<1:
        q=1

    if m2 < 1:

        m2 = 1.
        m1 = q*m2
        Mt = m1+m2


    phi_s = 0.0

    omega0 = 0.02
    f_min = omega0 / (np.pi * (m1 + m2) * lal.MTSUN_SI)
    distance = 1e6*lal.PC_SI
    delta_t = 1./4096.0#8192.0

    # Generate TD polatizations of SEOBNRv5HM

    """
    time_v5, modes_v5, hp_td_v5, hc_td_v5 = generate_v5PHM_waveform(m1, m2,
                                0.,0., chi1,
                                0.,0., chi2,
                                f_min,
                                iota_s, phi_s, distance,
                                delta_t, 'SEOBNRv5HM')




    # Generate TD polatizations of SEOBNRv5PHM

    time_v5p, modes_v5p, hp_td_v5p, hc_td_v5p = generate_v5PHM_waveform(m1, m2,
                                0.,0., chi1,
                                0.,0., chi2,
                                f_min,
                                iota_s, phi_s, distance,
                                delta_t, 'SEOBNRv5PHM')
    """


    time_v5, modes_v5, hp_td_v5, hc_td_v5 = generate_v5PHM_waveform(m1, m2,
                                0.,0., chi1,
                                0.,0., chi2,
                                f_min,
                                iota_s, phi_s, distance,
                                delta_t, 'IMRPhenomTPHM')


    time_v5p, modes_v5p, hp_td_v5p, hc_td_v5p = generate_v5PHM_waveform(m1, m2,
                                0.,0., chi1,
                                0.,0., chi2,
                                f_min,
                                iota_s, phi_s, distance,
                                delta_t, 'IMRPhenomTPHM')

    # Pad zeros
    N = max(len(hp_td_v5), len(hc_td_v5p))
    pad = int(2 ** (np.floor(np.log2(N)) + 2))
    hp_td_v5p.resize(pad)
    hc_td_v5p.resize(pad)
    hp_td_v5.resize(pad)
    hc_td_v5.resize(pad)


    # Perform the Fourier Transform
    hp_v5p = make_frequency_series(hp_td_v5p)
    hc_v5p = make_frequency_series(hc_td_v5p)


    hp_v5 = make_frequency_series(hp_td_v5)
    hc_v5 = make_frequency_series(hc_td_v5)


    amp_strain = abs(hp_v5p.data-1j*hc_v5p.data)
    idx_max = np.argmax(amp_strain)
    fpeak_v5p = hp_v5p.get_sample_frequencies()[idx_max]



    ampv5_strain = abs(hp_v5.data-1j*hc_v5.data)
    idx_max = np.argmax(ampv5_strain)
    fpeak_v5 = hp_v5.get_sample_frequencies()[idx_max]

    fpeak = max([fpeak_v5p,fpeak_v5])
    f_min = 1.35*fpeak

    # Generate PSD
    if f_min<10:
        f_low_phys = 10.
    else:
        f_low_phys = f_min


    f_high_phys = 2048.

    #"""
    try:
        #psd = aLIGOZeroDetHighPowerGWINC(len(hp_v5), hp_v5.delta_f, f_low_phys)
        psd = aLIGOZeroDetHighPower(len(hp_v5), hp_v5.delta_f, f_low_phys)

        # Compute match for hplus
        mm_hp = optimized_match(hp_v5,
              hp_v5p,
              psd,
              low_frequency_cutoff=f_low_phys,
              high_frequency_cutoff=f_high_phys
              )[0]


        # Compute match for hcross
        mm_hc = optimized_match(hc_v5,
              hc_v5p,
              psd,
              low_frequency_cutoff=f_low_phys,
              high_frequency_cutoff=f_high_phys
              )[0]

        # Take the mean
        mm_mean  = 1.-np.mean([mm_hp,mm_hc])

    except:

        print(
            f"Error for the following parameters: q = {q}, chi1 = {chi1}, chi2 = {chi2}, Mt = {Mt}, iota_s = {iota_s}"
        )
        mm_mean = -1
        pass
    #"""

    mm_mean=-1
    #print(m1,m2,chi1,chi2,iota_s,mm_mean)
    return (m1,m2, chi1, chi2, iota_s, mm_mean)


def process_one_case(input):

    m1, m2, chi1, chi2, iota = input

    m1,m2, chi1, chi2, iota, mm = mismatch_v5P_strain(
        m1,m2, chi1, chi2,iota
    )

    return  np.array([m1,m2, chi1, chi2, iota, mm])



if __name__ == "__main__":
    p = argparse.ArgumentParser(description="Compute mismatch")
    p.add_argument("--points", type=int, help="Number of points", default="5000")
    p.add_argument("--chi-max", type=float, help="Maximum spin", default="0.995")
    p.add_argument("--q-max", type=float, help="Maximum mass-ratio", default="100.0")
    p.add_argument(
        "--name", type=str, help="Name of the output file", default="mismatch"
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )
    p.add_argument("--n-cpu", type=int, help="Number of cores to use", default=64)
    p.add_argument(
        "--include-all",
        action="store_true",
        help="Include odd m modes close to equal mass",
        default=False,
    )

    p.add_argument(
        "--M-max", type=float, help="Maximum total mass", default="10.0"
    )
    p.add_argument(
        "--M-min", type=float, help="Maximum total mass", default="300.0"
    )

    args = p.parse_args()



    seed=150914
    N = args.points
    a_max = 0.998
    np.random.seed(seed)
    q = np.random.uniform(1., args.q_max, N)
    #m1 = np.random.uniform(args.M_min,args.M_max,N)
    np.random.seed(seed-1)
    mtotal = np.random.uniform(args.M_min, args.M_max, N)

    m1 = q/(1.+q)*mtotal
    m2 = 1./(1.+q)*mtotal

    np.random.seed(seed+999998)
    chi1arr = np.random.uniform(-a_max, a_max, N)
    np.random.seed(seed+999997)
    chi2arr = np.random.uniform(-a_max, a_max, N)
    np.random.seed(seed+999996)
    iota_arr = np.random.uniform(0,np.pi, N)

    lst = [ (m1_msun, m2_msun, s1z, s2z, iota) for m1_msun, m2_msun, s1z, s2z, iota in zip(m1,m2, chi1arr, chi2arr,iota_arr ) ]

    #print(f'Generated {args.points} random parameters')
    #print(lst)

    pool = Pool(args.n_cpu)
    all_means = pool.map(process_one_case, lst)
    all_means = np.array(all_means)

    np.savetxt(f"mismatch_aslimit_strain.dat", all_means)

    #print(f"Finalizing mismatch AS limit")
    ### Plots ###

    if args.plots:

        import matplotlib
        import matplotlib.pyplot as plt

        plt_dir = "./plots"
        Path(plt_dir).mkdir(parents=True, exist_ok=True)

        res_path = args.name
        params = np.genfromtxt(f"mismatch_aslimit_strain.dat")
        m1 = params[:, 0]
        m2 = params[:, 1]
        chi1 = params[:, 2]
        chi2 = params[:, 3]
        iota = params[:, 4]
        mm = params[:, 5]

        q = m1/m2
        mtotal = m1+m2
        nu = q / (1. + q) ** 2
        m1_dim = q / (1. + q)
        m2_dim = 1 / (1. + q)
        ap = (m1 * chi1 + m2 * chi2)/mtotal
        am = (m1 * chi1 - m2 * chi2)/mtotal

        # Histogram

        fig, ax = plt.subplots(figsize=(14,10),dpi=250)
        ax.hist(
            mm,
            bins=np.logspace(start=np.log10(1e-15), stop=np.log10(1.0), num=100),
            alpha=0.4,
            label='SEOBNRv5PHM - SEOBNRv5HM',
            histtype="step",
            fill=False
        )

        ax.axvline(np.median(mm), c="C0", ls="--")
        ax.legend(loc="best")
        ax.set_xlabel("$\mathcal{M}_{\mathrm{Max}}$",fontsize=30)
        ax.set_ylabel("Count",fontsize=30)
        ax.set_title(
            "$\mathcal{M}_{\mathrm{median}} = $" + f"{np.round(np.median(mm),10)}",fontsize=25
        )


        ax.set_xscale('log')

        size=30
        ax.tick_params(axis='x', which='major', pad=10,width=2,length=5,size=7, labelsize=size,direction='in')
        ax.tick_params(axis='x', which='minor', pad=10,width=2,length=5,size=7, labelsize=size,direction='in')
        ax.tick_params(axis='y', which='major', pad=10,width=2,length=5,size=7, labelsize=size,direction='in')
        ax.tick_params(axis='y', which='minor', pad=10,width=2,length=5,size=7, labelsize=size,direction='in')
        plt.tight_layout()
        plt.savefig(f"{plt_dir}/mm_hist_v5PHM_align_strain.png",bbox_inches="tight",dpi=300,)
        plt.close()

        # CDF
        fig, ax = plt.subplots(figsize=(14,10),dpi=250)
        ax.hist(
            mm,
            bins=np.logspace(start=np.log10(1e-15), stop=np.log10(1.0), num=100),
            alpha=0.4,
            label='SEOBNRv5PHM - SEOBNRv5HM',
            cumulative=True,
            density=True,
            histtype="step",
            lw=2,
        )




        ax.legend(loc="best")
        ax.set_xscale("log")
        ax.set_xlabel("$\mathcal{M}_{\mathrm{Max}}$")
        ax.set_title(
            "$\mathcal{M}_{\mathrm{median}} = $" + f"{np.round(np.median(mm),6)}"
        )

        ax.set_xscale('log')

        size=30
        ax.tick_params(axis='x', which='major', pad=10,width=2,length=5,size=7, labelsize=size,direction='in')
        ax.tick_params(axis='x', which='minor', pad=10,width=2,length=5,size=7, labelsize=size,direction='in')
        ax.tick_params(axis='y', which='major', pad=10,width=2,length=5,size=7, labelsize=size,direction='in')
        ax.tick_params(axis='y', which='minor', pad=10,width=2,length=5,size=7, labelsize=size,direction='in')
        plt.grid(True, which="both", ls=":")

        plt.tight_layout()
        plt.savefig(f"{plt_dir}/mm_cdf_v5PHM_align_limit_strain.png",bbox_inches="tight",dpi=300,)
        plt.close()



        # Mismatch across parameter space
        mm_s, q_s, ap_s = map(list, zip(*sorted(zip(mm, q, ap))))
        plt.scatter(
            q_s, ap_s, c=mm_s, linewidths=1, norm=matplotlib.colors.LogNorm()
        )
        plt.ylabel("$\chi_{\mathrm{eff}}$")
        plt.xlabel("$q$")
        cbar = plt.colorbar()
        cbar.set_label("$\mathcal{M}$")
        plt.savefig(
            f"{plt_dir}/mm_scatter_v5PHM_align_limit_strain.png",
            bbox_inches="tight",
            dpi=300,
        )
        plt.close()



"""
Additional utility functions.
"""

from typing import Any,Dict
import numpy as np

def frame_inv_amp(modes:Dict[Any,Any],ell_max:int=2,use_symm=True)->np.ndarray:
    """Compute the frame-invariant amplitude.
    By default, assumes that we have aligned-spin symmetry.

    Args:
        modes (Dict[Any,Any]): The dictionary of modes
        ell_max (int): The maximum l to use
        use_symm (bool): Assume up-down aligned symmetry

    Returns:
        np.ndarray: The frame-invariant ampltiude time series
    """
    total = 0.0

    for mode in modes.keys():
        ell,m = mode
        if ell>ell_max:
            continue
        if use_symm:
            total+=2*np.abs(modes[mode])**2
        else:
            total+=np.abs(modes[mode])**2
    return np.sqrt(total)



"""
Contains functions to evaluate the MR ansatze. 
See Eq. (56) and following in https://dcc.ligo.org/DocDB/0186/T2300060/001/SEOBNRv5HM.pdf.
"""

import numpy as np
from ..fits.EOB_fits import *
from ..utils.utils_precession_opt import compute_omegalm_P_frame


def compute_MR_mode_free(
    t, m1, m2, chi1, chi2, attach_params, ell, m, fits_dict,f_nyquist,lmax_nyquist,t_match=0, phi_match=0, qnm_rotation=0.
):
    """
    Evaluate the MR ansatze. See Eqs.(56, 57, 58) in https://dcc.ligo.org/DocDB/0186/T2300060/001/SEOBNRv5HM.pdf.

    Args:
        t (np.ndarray): ringdown time array
        m1 (float): mass of the primary
        m2 (float): mass of the secondary
        chi1 (float): dimensionless spin of the primary
        chi2 (float): dimensionless spin of the secondary
        attach_params (dict): dictionary containing quantities needed to compute the
            MR ansatz (input value fits, final mass and spin)
        ell (int): ell index of the desired mode
        m (int): m index of the desired mode
        t (np.ndarray): inspiral time array
        fits_dict (dict): dictionary of fit coefficients in the ringdown anzatz
        t_match (float): ringdown time at which the merger-ringdown waveform is started
        phi_match (float): value of the phase at t_match
        qnm_rotation (float): Factor rotating the QNM mode frequency in the co-precessing frame (Eq. 33 of Hamilton et al.)

    Returns:
        np.ndarray: the merger-ringdown waveform

    """
    # Step 1 - use the NR fits for amplitude and frequency at attachment time
    eta = m1 * m2 / (m1 + m2) ** 2

    nra = attach_params["amp"]
    nraDot = attach_params["damp"]
    nromega = attach_params["omega"]

    # Step 2 - compute the  QNMs
    final_mass = attach_params["final_mass"]
    final_spin = attach_params["final_spin"]
    omega_complex = compute_QNM(ell, m, 0, final_spin, final_mass).conjugate()
    omega_complex = compute_omegalm_P_frame(omega_complex, m, qnm_rotation)

    # Sanity check on nyquist frequency
    if np.real(omega_complex) > 2*np.pi*f_nyquist and lmax_nyquist >= ell:
        raise ValueError(
                f"Internal function call failed: Input domain error. Ringdown frequency of ({ell},{m}) mode greater than maximum frequency from Nyquist theorem. Decrease the time spacing to resolve this mode."
            )
    sigmaR = -np.imag(omega_complex)
    sigmaI = -np.real(omega_complex)
    # Step 3 - use the fits for the free coefficients in the RD anzatse
    c1f = fits_dict['c1f'][(ell,m)]
    c2f = fits_dict['c2f'][(ell,m)]
    # Ensure that the ampitude of the (2,2) mode has a maximum at t_match
    if ell == 2 and m == 2:
        if sigmaR > 2.0 * c1f * np.tanh(c2f):
            c1f = sigmaR / (2.0 * np.tanh(c2f))

    d1f = fits_dict['d1f'][(ell,m)]
    d2f = fits_dict['d2f'][(ell,m)]
    # Step 4 - compute the constrainted coefficients
    c1c = EOBCalculateRDAmplitudeConstraintedCoefficient1(
        c1f, c2f, sigmaR, nra, nraDot, eta
    )
    c2c = EOBCalculateRDAmplitudeConstraintedCoefficient2(
        c1f, c2f, sigmaR, nra, nraDot, eta
    )
    d1c = EOBCalculateRDPhaseConstraintedCoefficient1(d1f, d2f, sigmaI, nromega)



    # Step 5 - assemble the amplitude and phase
    # First the ansatze part
    Alm = c1c * np.tanh(c1f * (t - t_match) + c2f) + c2c
    philm = phi_match - d1c * np.log(
        (1 + d2f * np.exp(-d1f * (t - t_match))) / (1 + d2f)
    )

    test_omega = -np.real(omega_complex) + 1j * np.imag(omega_complex)
    # hlm = eta*Alm*np.exp(1j*philm)*np.exp(1j*omega_complex*(t-t_match))
    hlm = eta * Alm * np.exp(1j * philm) * np.exp(1j * test_omega * (t - t_match))
    return hlm



import numpy as np
from skopt.space import Space
from skopt.sampler import Lhs


def parameters_random(
    N: int,
    qmin: float,
    qmax: float,
    chi1min: float,
    chi1max: float,
    chi2min: float,
    chi2max: float,
    random_state: int = None,
):
    """
    Generate a Latin Hypercube sampling for q, chi1, chi2

    Parameters
    ----------
    N:
        Number of parameters to generate
    qmin, qmax:
        Bounds on the q parameter
    chi1min, chi1max, chi2min, chi2max:
        Bounds on the chi parameter

    Return parameters of the waveforms
    """
    space = Space([(qmin, qmax), (chi1min, chi1max), (chi2min, chi2max)])
    lhs = Lhs(
        lhs_type="classic", criterion="maximin", iterations=1000
    )  # Latin hypercube sampling
    q = []
    chiA = []
    chiB = []
    if N > 0:
        rand = lhs.generate(space.dimensions, N, random_state = random_state)
        for i in range(N):
            q.append(rand[i][0])
            chiA.append(rand[i][1])
            chiB.append(rand[i][2])
    return (
        np.array(q),
        np.array(chiA),
        np.array(chiB),
    )


def parameters_random_2D(
    N: int,
    qmin: float,
    qmax: float,
    chi1min: float,
    chi1max: float,
    random_state: int = None,
):
    """
    Generate a Latin Hypercube sampling for q, chi1 (for NRHybSur2dq15)

    Parameters
    ----------
    N:
        Number of parameters to generate
    qmin, qmax:
        Bounds on the q parameter
    chi1min, chi1max:
        Bounds on the chi parameter

    Return parameters of the waveforms
    """
    space = Space([(qmin, qmax), (chi1min, chi1max)])
    lhs = Lhs(
        lhs_type="classic", criterion="maximin", iterations=1000
    )  # Latin hypercube sampling
    q = []
    chiA = []
    if N > 0:
        rand = lhs.generate(space.dimensions, N, random_state = random_state)
        for i in range(N):
            q.append(rand[i][0])
            chiA.append(rand[i][1])
    return (
        np.array(q),
        np.array(chiA),
    )

def q_to_nu(q):
    return q/(1+q)**2

def nu_to_q(nu):
    delta_m = np.sqrt(np.abs(1 - 4 * nu))
    return ((1 - 2 * nu) + delta_m) / (2 * nu)


def parameters_random_fast(
    N: int,
    qmin: float,
    qmax: float,
    chi1min: float,
    chi1max: float,
    chi2min: float,
    chi2max: float,
    random_state=1
):
    """
    Generate random parameters for q, chi1, chi2

    Parameters
    ----------
    N:
        Number of parameters to generate
    qmin, qmax:
        Bounds on the q parameter
    chi1min, chi1max, chi2min, chi2max:
        Bounds on the chi parameter

    Return parameters of the waveforms
    """
    np.random.seed(random_state)
    #q = np.random.uniform(qmin, qmax, N)
    nu = np.random.uniform(q_to_nu(qmin),q_to_nu(qmax),N)
    q = nu_to_q(nu)
    chiA = np.random.uniform(chi1min, chi1max, N)
    chiB = np.random.uniform(chi2min, chi2max, N)

    return (
        np.array(q),
        np.array(chiA),
        np.array(chiB),
    )


def parameters_random_2D_fast(
    N: int,
    qmin: float,
    qmax: float,
    chi1min: float,
    chi1max: float,
    random_state=1,
):
    """
    Generate random parameters for q, chi1 (for NRHybSur2dq15)

    Parameters
    ----------
    N:
        Number of parameters to generate
    qmin, qmax:
        Bounds on the q parameter
    chi1min, chi1max:
        Bounds on the chi parameter

    Return parameters of the waveforms
    """
    np.random.seed(random_state)

    #q = np.random.uniform(qmin, qmax, N)
    nu = np.random.uniform(q_to_nu(qmin),q_to_nu(qmax),N)
    q = nu_to_q(nu)
    chiA = np.random.uniform(chi1min, chi1max, N)

    return (
        np.array(q),
        np.array(chiA),
    )


cimport cython
from pyseobnr.eob.utils.containers cimport EOBParams
from pyseobnr.eob.waveform.waveform cimport RR_force,  RadiationReactionForce
from pyseobnr.eob.hamiltonian.Hamiltonian_v5PHM_C cimport Hamiltonian_v5PHM_C


cpdef get_rhs_prec(double t,double[::1] z,Hamiltonian_v5PHM_C H,RadiationReactionForce RR,
    double m_1,double m_2,EOBParams params)


from functools import lru_cache
from typing import Union

import lal
import numpy as np
from pycbc.filter import make_frequency_series
from pycbc.filter.matchedfilter import match
from pycbc.types import FrequencySeries, TimeSeries
from pycbc.psd.analytical import aLIGOZeroDetHighPowerGWINC, EinsteinTelescopeP1600143, CosmicExplorerP1600143

from scipy.signal import argrelmax
from waveform_tools.mismatch.unfaithfulness import generate_waveform
from waveform_tools.mismatch.waveform_parameters import waveform_params
from waveform_tools.mismatch.auxillary_funcs import Ylm
import matplotlib.pyplot as plt


def _planckWindowFunction(t, tol=0.005):
    """f(t) = 0 if t < tol
    = 1 if t > 1 - tol
    = 1/(1 + exp(z)) otherwise, where z = 1/t - 1/(1-t)"""
    safe = (t > tol) * (t < (1.0 - tol))
    # temporarily set unsafe times to 0.5 to avoid dividing by 0
    safeT = safe * t + (1.0 - safe) * 0.5
    safeZ = 1.0 / safeT - 1.0 / (1.0 - safeT)
    return safe * 1.0 / (1.0 + np.exp(safeZ)) + (t >= (1.0 - tol))


def window_function(t, t_s=0.0, t_e=1.0, rolloff=False):
    """Return a Planck window function, evaluated at an array of times t:
    When t <= t_s, f(t) = 0
    When t >= t_e, f(t) = 1
    Otherwise, 0 <= f(t) <= 1, where f is monotonically increasing.

    If rolloff=True, reverse the window so that f(t_s) = 1, f(t_e) = 0.

    This is a modified version of the code from gwtools by C.Galley
    """
    if rolloff:
        return window_function(-t, t_s=-t_e, t_e=-t_s)

    # Rescale the time so that the time interval of interest is always
    # 0 to 1
    scaled_t = (t - t_s) / (t_e - t_s)
    return _planckWindowFunction(scaled_t)


def taper_waveform(ts, h, rollon_end=4, ending_time=35):
    maxs = argrelmax(h, order=20)[0]
    # Roll-on interval
    t_l1 = ts[0]
    t_l2 = ts[maxs[rollon_end]]

    # Roll-off interval
    t_h1 = ts[-1] - ending_time
    t_h2 = ts[-1]

    # Roll-on window
    window1 = window_function(ts, t_s=t_l1, t_e=t_l2)
    # Roll-off window
    window2 = window_function(ts, t_s=t_h1, t_e=t_h2, rolloff=True)
    return h * window1 * window2


def taper_pycbc_series(time_series, rollon_end=4):
    ts = time_series.sample_times.data
    h = time_series.numpy()
    return TimeSeries(
        taper_waveform(ts, h, rollon_end=rollon_end), delta_t=time_series.delta_t
    )


def get_padded_length(h):
    N = len(h)
    return int(2 ** (np.floor(np.log2(N)) + 2))


def condition_waveform(t, h, n=None, convert=True):
    # Taper
    h = taper_waveform(t, h)
    # Pad

    if n is None:
        n_pad = get_padded_length(h)
    else:
        n_pad = int(2 ** (np.floor(np.log2(n)) + 2))
    h.resize(n_pad)
    if convert:
        h = TimeSeries(h, delta_t=np.diff(t)[0])
    return h


def condition_pycbc_series(h, n=None):
    # Taper
    h = taper_pycbc_series(h)
    # Pad

    if n is None:
        n_pad = get_padded_length(h)
    else:
        n_pad = int(2 ** (np.floor(np.log2(n)) + 2))
    h.resize(n_pad)
    return h


@lru_cache(maxsize=128)
def generate_psd(
    length: int, delta_f: float, flow: float, psd_type: str = "aLIGO"
) -> FrequencySeries:
    """A memoized version of different PSDs
    Available:
    'aLIGO' - aLIGOZeroDetHighPowerGWINC
    'flat' - unity PSD


    Args:
        length (int): length of the PSD to generate
        delta_f (float): frequency spacing
        flow (float): low frequency cutoff

    Returns:
        pt.FrequencySeries: the PSD
    """
    if psd_type == "aLIGO":
        return aLIGOZeroDetHighPowerGWINC(length, delta_f, flow)
    elif psd_type == "ET":
        return EinsteinTelescopeP1600143(length, delta_f, flow)
    elif psd_type == "CE":
        return CosmicExplorerP1600143(length, delta_f, flow)
    elif psd_type == "flat":
        return FrequencySeries(np.ones(length), delta_f=delta_f)
    else:
        raise NotImplementedError


def fast_unfaithfulness_mode_by_mode(
    h1: Union[TimeSeries, FrequencySeries],
    h2: Union[TimeSeries, FrequencySeries],
    f_low: float,
    Ms: np.ndarray,
    fmin: float = 10.0,
    fmax: float = 2048.0,
    verbose: bool = False,
    psd_t: str = "aLIGO",
):
    """Compute the simple match (optimisation over time and phase) between
    two waveforms. The waveforms are _assumed to be in geometric units_.
    In time-domain the waveforms are assumed to be already conditioned.
    The spacing/length in time and frequency is assumed to be already consistent.
    For example:
    - if both waveforms are in TD, time spacing is the same
    - if both waveforms are in FD, frequency spacing *and length* are the same
    - if one is in TD and one in FD, it is assumed that calling an FFT on the TD
    waveform will give the right frequency spacing and length.

    The mismatch integral will be done as follows:
    - between f_low and fmax if f_low>fmin
    - between fmin and fmax if f_low<fmin

    Args:
        h1 (Union[TimeSeries, FrequencySeries]): The first waveform in geometric units
        h2 (Union[TimeSeries, FrequencySeries]): The second waveform in geometric units
        f_low (float): The minimum of the strarting frequencies of the 2 waveforms, in geometric units
        Ms (np.ndarray): Masses to use, in solar masses
        fmin (float, optional): Minimum limit of mismatch integral if possible [Hz]. Defaults to 10.0.
        fmax (float, optional): Maximum limit of mismatch integral [Hz]. Defaults to 2048.0.
        verbose (bool, optional): Print extra info. Defaults to False.
        psd_t (str, optional): The type of PSD to use. Defaults to "aLIGO".

    Returns:
        np.ndarray: The array of mismatches
    """

    # Distance scale: feducial but chosen so that the amplitude of the modes
    # is not too far in order of magnitude from the PSD
    dist = 1.0e7 * lal.PC_SI / lal.C_SI
    matches = np.zeros(len(Ms))
    scale_fac = 1.0
    # print(f"f_low={f_low}")
    for i, M in enumerate(Ms):
        # Mass in seconds
        Mt = M * lal.MTSUN_SI

        # physical unit
        f_high_phys = fmax
        f_low_phys = f_low / Mt
        if f_low_phys < fmin:
            f_low_phys = fmin
        if verbose:
            print("\n\n f_low = ", f_low_phys, " Hz, f_high = ", f_high_phys, " Hz")

        if i == 0:
            # First time we are computing things
            # Rescale signal and template to physical units as appropriate for time
            # or frequency domain waveforms
            dist_scale = M * lal.MRSUN_SI * M * lal.MTSUN_SI / dist
            if isinstance(h1, TimeSeries):

                dtM = h1.delta_t * Mt
                x1 = TimeSeries(h1 * Mt / dist, delta_t=dtM)
                h1tilde = make_frequency_series(x1)
            else:
                h1tilde = FrequencySeries(
                    h1.numpy() * dist_scale, delta_f=h1.delta_f / (M * lal.MTSUN_SI)
                )

            if isinstance(h2, TimeSeries):
                dtM = h2.delta_t * Mt
                x2 = TimeSeries(h2 * Mt / dist, delta_t=dtM)
                h2tilde = make_frequency_series(x2)
            else:
                h2tilde = FrequencySeries(
                    h2.numpy() * dist_scale, delta_f=h2.delta_f / (M * lal.MTSUN_SI)
                )

        else:
            # We now have signal and template in frequency domain. Just rescale
            # correctly with the new total mass. Note that the amplitude scaling
            # is not done since it's irrelevant to the mismatch
            scale_fac = 1.0 * M / Ms[i - 1]
            h1old = h1tilde
            h2old = h2tilde
            h1tilde = FrequencySeries(h1old.data, delta_f=h1old.delta_f / scale_fac)
            h2tilde = FrequencySeries(h2old.data, delta_f=h2old.delta_f / scale_fac)

        psd = generate_psd(len(h2tilde), h2tilde.delta_f, f_low_phys, psd_type=psd_t)

        matches[i] = match(
            h2tilde,
            h1tilde,
            psd=psd,
            low_frequency_cutoff=f_low_phys,
            high_frequency_cutoff=f_high_phys,
        )[0]

    return matches


def generate_dominant_mode_pol_LAL(params: waveform_params, f_max: float = 2048):
    # Generate the dominant mode
    hp, hc = generate_waveform(params, f_max=f_max)
    M = params.m1 + params.m2
    # For convenience we now make everything be in geometric units
    if params.domain == "TD":
        # Rescale the amplitude and time array
        amp_prefac = M * lal.MRSUN_SI / params.distance * np.abs(Ylm(2, 2, 0, 0))
        time_prefac = 1 / (M * lal.MTSUN_SI)
        return TimeSeries(hp / amp_prefac, delta_t=hp.delta_t * time_prefac)
        #
    elif params.domain == "FD":
        # Rescale the amplitude and frequency array
        # Amplitude rescaling prefactor. Note the extra factors that come from FFT!
        amp_prefac = (
            M
            * lal.MRSUN_SI
            * M
            * lal.MTSUN_SI
            / params.distance
            * np.abs(Ylm(2, 2, 0, 0))
        )
        freq_prefac = M * lal.MTSUN_SI
        return FrequencySeries(hp / amp_prefac, delta_f=hp.delta_f * freq_prefac)



from typing import Any, Dict

import numpy as np
from scipy.signal import argrelmax


def one_peak_test(t: np.ndarray, h: np.ndarray, interval_end: float) -> None:
    """Check that a quantity has only 1 peak until
    interval_end

    Args:
        t (np.ndarray): The time array
        h (np.ndarray): The quantity as function of tim
        interval_end (float): Last time to consider
    """
    times = np.where(t <= interval_end)
    t = 1 * t[times]
    h = 1 * h[times]
    maxs = argrelmax(h)
    global_max = np.argmax(h)
    assert len(maxs[0]) == 1, "Should only have 1 max in this quantity"
    assert maxs[0] == global_max, f"Local max is not global max,{maxs[0]},{global_max}"


def monotonic_quantity_test(
    t: np.ndarray, h: np.ndarray, interval_end: float, threshold: float = 1e-10
) -> None:
    """Check that thie give quantity is monotonic until interval_end,
    with small deviations allowed, quantified by threshold parameter

    Args:
        t (np.ndarray): The time array
        h (np.ndarray): Quantity as a function of tim
        interval_end (float): Last time to consider
        threshold (float, optional): Allowed deviation from monotonicity. Defaults to 1e-10.
    """
    times = np.where(t <= interval_end)
    t = 1 * t[times]
    h = 1 * h[times]
    forward_diffs = np.diff(h)

    # Allow for small inaccuracies in monotonicity
    conds = (forward_diffs > 0) | (np.abs(forward_diffs) < threshold)

    assert np.alltrue(conds), "The quantity has to be strictly monotonic"


def amplitude_hierarchy_test(t: float, modes: Dict[Any, Any]) -> None:
    """Check that the amplitude of the subdominant modes is
    less than the amplitude of the (2,2) mode until the peak
    of the (2,2) modes

    Args:
        t (np.array): Times
        modes (Dict): The dictionary of modes
    """
    h22 = modes["2,2"]
    amp22 = np.abs(h22)
    interval_end = t[np.argmax(amp22)]
    times = np.where(t <= interval_end)[0]
    amp22 = amp22[times]
    for ell_m in modes.keys():
        if ell_m == "2,2":
            continue
        current_mode = modes[ell_m]
        amp_mode = np.abs(current_mode)
        amp_mode = amp_mode[times]
        assert np.alltrue(
            amp_mode < amp22
        ), f"Amplitude of mode {ell_m} should be less than the (2,2) mode before merger"




import argparse
import os

from bilby.core.utils import check_directory_exists_and_if_not_mkdir
from jinja2 import Environment, FileSystemLoader

this_folder = os.path.abspath(os.path.dirname(__file__))
script_home = os.path.join(this_folder, "../../")


if __name__ == "__main__":
    p = argparse.ArgumentParser()
    p.add_argument(
        "--n-cores",
        type=int,
        help="Number of cores to use. Should match number of tasks per node",
        default=64,
    )
    p.add_argument("--queue", type=str, help="The queue to use", default="syntrofos")
    p.add_argument(
        "--run-time",
        type=int,
        help="Number of hours to ask from the cluster",
        default=23,
    )
    p.add_argument(
        "--env", type=str, help="The path to the python env one wishes to use"
    )
    p.add_argument(
        "--test-dir",
        type=str,
        help="Name of directory where tests will be created",
        default="./",
    )
    p.add_argument(
        "--wrapper-path", type=str, help="The path to the wrapper, including name"
    )
    args = p.parse_args()
    # Create submit files for the different tests
    file_loader = FileSystemLoader(f"{this_folder}/templates/")
    env = Environment(loader=file_loader)
    template = env.get_template("slurm.jinja")

    check_directory_exists_and_if_not_mkdir(args.test_dir)
    os.chdir(args.test_dir)
    # Mismatch with NR
    cmd = f"""python {script_home}/auxiliary/sanity_checks/NR_mismatches.py --cases-file /work/lpompili/csv_files/spinning_calibs.csv"""
    prog = template.render(
        job_name="NR_mismatch",
        label="sanity_check",
        ncores=args.n_cores,
        queue=args.queue,
        time=args.run_time,
        exec_script=cmd,
        env=args.env,
    )
    with open("NR_mismatches.sh", "w") as fw:
        fw.write(prog)
    # Mismatch with NRSur3dq8
    cmd = f"""python {script_home}/auxiliary/sanity_checks/NRSur_matches.py --points 5000 --chi-max 0.9 --q-max 8.0 --name mm_3dq8_90.dat --model-name NRHybSur3dq8 --approximant SEOBNRv5HM"""
    prog = template.render(
        job_name="NRSur3dq8_mismatch",
        label="sanity_check",
        ncores=args.n_cores,
        queue=args.queue,
        time=args.run_time,
        exec_script=cmd,
        env=args.env,
    )
    with open("NRSur3dq8_mismatches.sh", "w") as fw:
        fw.write(prog)

    # Mismatch with NRSur2dq15
    cmd = f"""python {script_home}auxiliary/sanity_checks/NRSur_matches.py --points 5000 --chi-max 0.6 --q-max 15.0 --name mm_2dq15_60.dat --model-name NRHybSur2dq15 --approximant SEOBNRv5HM"""
    prog = template.render(
        job_name="NRSur2dq15_mismatch",
        label="sanity_check",
        ncores=args.n_cores,
        queue=args.queue,
        time=args.run_time,
        exec_script=cmd,
        env=args.env,
    )
    with open("NRSur2dq15_mismatches.sh", "w") as fw:
        fw.write(prog)

    # Mismatch with SEOBNRv4HM - q<=10
    cmd = f"""python {script_home}auxiliary/sanity_checks/EOB_matches.py --points 10000 --chi-max 0.998 --q-max 10 --plots --model-1-name SEOBNRv4HM --model-2-name SEOBNRv5HM --n-cpu 64 --name comp"""
    prog = template.render(
        job_name="SEOBNRv4HM_mismatch",
        label="sanity_check",
        ncores=args.n_cores,
        queue=args.queue,
        time=args.run_time,
        exec_script=cmd,
        env=args.env,
    )
    with open("SEOBNv4HM_mismatches.sh", "w") as fw:
        fw.write(prog)

    # Mismatch with SEOBNRv4HM - q>10
    cmd = f"""python {script_home}auxiliary/sanity_checks/EOB_matches.py --points 3000 --q-min 10.0 --chi-max 0.96 --q-max 100 --plots --model-1-name SEOBNRv4HM --model-2-name SEOBNRv5HM --n-cpu 64 --name ext"""
    prog = template.render(
        job_name="SEOBNRv4HM_mismatch_ext",
        label="sanity_check",
        ncores=args.n_cores,
        queue=args.queue,
        time=args.run_time,
        exec_script=cmd,
        env=args.env,
    )
    with open("SEOBNv4HM_mismatches_ext.sh", "w") as fw:
        fw.write(prog)



    # Amplitude monotonicity
    cmd = f"""python {script_home}auxiliary/sanity_checks/monotonicity_test.py --points 100000  --quantity amplitude"""
    prog = template.render(
        job_name="amplitude_monotonicity",
        label="sanity_check",
        ncores=args.n_cores,
        queue=args.queue,
        time=args.run_time,
        exec_script=cmd,
        env=args.env,
    )
    with open("amp_monotonicity.sh", "w") as fw:
        fw.write(prog)

    # Amplitude monotonicity (more cases for high q)
    cmd = f"""python {script_home}auxiliary/sanity_checks/monotonicity_test.py --q-min 70.0 --points 500000 --name monotonicity_test_high_q  --quantity amplitude"""
    prog = template.render(
        job_name="amplitude_monotonicity_high_q",
        label="sanity_check",
        ncores=args.n_cores,
        queue=args.queue,
        time=args.run_time,
        exec_script=cmd,
        env=args.env,
    )
    with open("amp_monotonicity_high_q.sh", "w") as fw:
        fw.write(prog)

    # Frequency mononicity
    cmd = f"""python {script_home}auxiliary/sanity_checks/monotonicity_test.py --points 100000  --quantity frequency"""
    prog = template.render(
        job_name="frequency_monotonicity",
        label="sanity_check",
        ncores=args.n_cores,
        queue=args.queue,
        time=args.run_time,
        exec_script=cmd,
        env=args.env,
    )
    with open("freq_monotonicity.sh", "w") as fw:
        fw.write(prog)

    # Amplitude hierarchy
    cmd = f"""python {script_home}auxiliary/sanity_checks/amplitude_hierarchy_test.py --points 100000 --q-max 100 --chi-max 0.998"""
    prog = template.render(
        job_name="amplitude_hierarchy",
        label="sanity_check",
        ncores=args.n_cores,
        queue=args.queue,
        time=args.run_time,
        exec_script=cmd,
        env=args.env,
    )
    with open("amplitude_hierarchy.sh", "w") as fw:
        fw.write(prog)

    # Smoothness varying q, chi
    cmd = f"""python {script_home}auxiliary/sanity_checks/smoothness_q_chi.py"""
    prog = template.render(
        job_name="smoothness_plots",
        label="sanity_check",
        ncores=args.n_cores,
        queue=args.queue,
        time=args.run_time,
        exec_script=cmd,
        env=args.env,
    )
    with open("smoothness_plots.sh", "w") as fw:
        fw.write(prog)



import numpy as np
import argparse
import os
import sys
from schwimmbad import JoblibPool
from bilby.core.utils import check_directory_exists_and_if_not_mkdir

sys.path.append( os.path.join( os.path.abspath(os.path.dirname(__file__)), "../../calibration" ) )
sys.path.append( os.path.join( os.path.abspath(os.path.dirname(__file__)), "../../toys/hamiltonian_prototype" ) )

from compute_hlms import Kerr_ISCO_mod
from models import *
from metrics import *
from parameters import parameters_random

from SEOBNRv5_wrapper import SEOBNRv5_wrapper


def reference_time(q, chi1, chi2):
    """
    Generate a SEOBNRv5 waveform with parameters (q, chi1, chi2) and compute
    the reference point for the MR attachment (r_ISCO with modified final spin)
    and the last r of the dynamics rend. 

    One should always have r_ISCO > rend
    """
    omega0 = 0.02

    model = SEOBNRv5_wrapper(
        q,
        chi1,
        chi2,
        omega0,
    )
    
    m1 = q/(1+q)
    m2 = 1/(1 + q)
    r_ISCO, L_ISCO = Kerr_ISCO_mod(chi1, chi2, m1, m2)
    rend = model.dynamics[:,1][-1]

    return (
        q,
        chi1,
        chi2,
        rend,
        r_ISCO,
    )


def process_one_case(input):
    q, chi1, chi2 = input
    q, chi1, chi2, rend, r_ISCO = reference_time(q, chi1, chi2)
    return np.array([q, chi1, chi2, rend, r_ISCO])


if __name__ == "__main__":

    p = argparse.ArgumentParser(
        description="Check reference time smoothness across parameter space"
    )
    p.add_argument(
        "--points", type=int, help="Number of points", default="10000"
    )
    p.add_argument(
        "--q-max", type=float, help="Upper limit on q", default="100.0"
    )
    p.add_argument(
        "--chi-max", type=float, help="Upper limit on |chi|", default="0.995"
    )
    p.add_argument(
        "--name", type=str, help="Name of the output file", default="reference_time.dat"
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )
    args = p.parse_args()

    qarr, chi1arr, chi2arr = parameters_random(args.points, 1.0, args.q_max, -args.chi_max, args.chi_max, -args.chi_max, args.chi_max)

    lst = [(a, b, c) for a, b, c in zip(qarr, chi1arr, chi2arr)]

    with JoblibPool(24) as pool:
        all_means = pool.map(process_one_case, lst)

    all_means = np.array(all_means)

    np.savetxt(args.name, all_means)

    ### Plots ###

    if args.plots:
        
        import matplotlib
        import matplotlib.pyplot as plt
        matplotlib.rcParams['text.usetex'] = True

        plt_dir = './plots'
        check_directory_exists_and_if_not_mkdir(plt_dir)

        res_path = args.name
        res = np.loadtxt(res_path)

        q = res[:,0]
        chi1 = res[:,1]
        chi2 = res[:,2]
        rend = res[:,3]
        r_ISCO = res[:,4]

        nu = q/(1+q)**2
        m1 = q/(1+q)
        m2 = 1/(1 + q)
        ap = m1*chi1+m2*chi2
        am = m1*chi1-m2*chi2

        # End r across parameter space - (q, chieff)
        plt.scatter(q, ap, c=rend,linewidths=1)
        plt.ylabel('$\chi_{\mathrm{eff}}$')
        plt.xlabel('$q$')
        cbar=plt.colorbar()
        cbar.set_label('$r_{\mathrm{end}} (\Omega_0 = 0.015)$')
        plt.savefig(f'{plt_dir}/r_end.png', bbox_inches = 'tight', dpi = 300)
        plt.close()

        # rend - rISCO across parameter space - (q, chieff)
        plt.scatter(q, ap, c=rend - r_ISCO,linewidths=1)
        plt.ylabel('$\chi_{\mathrm{eff}}$')
        plt.xlabel('$q$')
        cbar=plt.colorbar()
        cbar.set_label('$r_{\mathrm{end}} - r_{\mathrm{ISCO}}$')
        plt.savefig(f'{plt_dir}/rend_rISCO.png', bbox_inches = 'tight', dpi = 300)
        plt.close()



"""
Contains functions that are used in evaluating different waveform fits. Also includes computation of NQCs
"""

from typing import Any, Dict

import numpy as np
import qnm
from scipy.interpolate import InterpolatedUnivariateSpline

# Pre-cache the QNM interpolants
modes_qnm = [(2, 2), (2, 1), (3, 3), (3, 2), (4, 4), (4, 3), (5, 5)]
qnm_cache = {}
qnm_cache_mm = {}
for mode in modes_qnm:
    ell, m = mode
    qnm_cache[ell, m] = qnm.modes_cache(s=-2, l=ell, m=m, n=0)
    qnm_cache_mm[ell, m] = qnm.modes_cache(s=-2, l=ell, m=-m, n=0)


def compute_QNM(ell: int, m: int, n: int, af: float, Mf: float):
    """Return the complex QNM frequencies by interpolating
    existing solutions. Note that for negative spins we use
    the _positive_ spin negative m call internally

    Args:
        ell (int): ell of the desired QNM
        m (int): m of the desired QNM
        n (int): overtone. Only n=0 is supported
        af (float): final spin of the BH
        Mf (float): final mass of the BH

    Returns:
        np.complex128: The complex QNM frequency.
        The frequencies are such that Re(omega)>0 and Im(omega)<0.
    """
    if af > 0:
        omega, _, _ = qnm_cache[(ell, m)](a=af, interp_only=True)
    else:
        omega, _, _ = qnm_cache_mm[(ell, m)](a=np.abs(af), interp_only=True)
    return omega / Mf


def EOBCalculateRDAmplitudeConstraintedCoefficient1(
    c1f: float, c2f: float, sigmaR: float, amp: float, damp: float, eta: float
):
    """
    Computes c1c coefficient in MR ansatze (Eq 59 in https://dcc.ligo.org/DocDB/0186/T2300060/001/SEOBNRv5HM.pdf).

    Args:
        c1f (float): value of c1f
        c2f (float): value of c1f
        sigmaR (float): real part of the QNM frequency
        amp (float): waveform amplitude at attachment time
        damp (float): waveform amplitude's first derivative at attachment time
        eta (float): reduced mass ratio

    Returns:
        float: The value of c1c
    """
    c1c = 1 / (c1f * eta) * (damp - sigmaR * amp) * np.cosh(c2f) ** 2
    return c1c


def EOBCalculateRDAmplitudeConstraintedCoefficient2(
    c1f: float, c2f: float, sigmaR: float, amp: float, damp: float, eta: float
):
    """
    Computes c2c coefficient in MR ansatze (Eq 60 in https://dcc.ligo.org/DocDB/0186/T2300060/001/SEOBNRv5HM.pdf).

    Args:
        c1f (float): value of c1f
        c2f (float): value of c1f
        sigmaR (float): real part of the QNM frequency
        amp (float): waveform amplitude at attachment time
        damp (float): waveform amplitude's first derivative at attachment time
        eta (float): reduced mass ratio

    Returns:
        float: The value of c2c
    """
    c2c = amp / eta - 1 / (c1f * eta) * (damp - sigmaR * amp) * np.cosh(c2f) * np.sinh(
        c2f
    )
    return c2c


def EOBCalculateRDPhaseConstraintedCoefficient1(
    d1f: float, d2f: float, sigmaI: float, omega: float
):
    """
    Computes d1c coefficient in MR ansatze (Eq 61 in https://dcc.ligo.org/DocDB/0186/T2300060/001/SEOBNRv5HM.pdf).

    Args:
        d1f (float): value of d1f
        d2f (float): value of d1f
        sigmaI (float): imaginary part of the QNM frequency
        omega (float): waveform frequency at attachment time

    Returns:
        float: The value of c2c
    """
    d1c = (omega - sigmaI) * (1 + d2f) / (d1f * d2f)
    return d1c


def EOBCalculateNQCCoefficients_freeattach(
    amplitude,
    phase,
    r,
    pr,
    omega_orb,
    ell,
    m,
    time_peak: float,
    time,
    m1: float,
    m2: float,
    chi1: float,
    chi2: float,
    nrDeltaT: float,
    fits_dict: Dict[str, Any],
):
    """
    Computes the NQC coefficients (see discussion in Sec. II A, around Eq(35), in https://dcc.ligo.org/DocDB/0186/T2300060/001/SEOBNRv5HM.pdf)
    This is just like the SEOBNRv4HM function but allows nrDeltaT to be passed in, instead of
    internally calculated.

    Args:
        amplitude (np.ndarray): amplitude of the relevant mode
        phase (np.ndarray): phase of the relevant mode
        r (np.ndarray): r along the dynamics
        pr (np.ndarray): pr along the dynamics
        omega_orb (np.ndarray): omega_orb along the dynamics
        ell (int): ell index of the relevant mode
        m (int): m index of the relevant mode
        time_peak (float): "reference" time with respect to which the attachment 
            is defined, corresponding to t_ISCO for SEOBRNRv5HM (see Eq. 41 of https://dcc.ligo.org/DocDB/0186/T2300060/001/SEOBNRv5HM.pdf)
        time (np.ndarray): time array
        m1 (float): primary mass
        m2 (float): secondary mass
        chi1 (float): primary mass
        chi2 (float): primary mass
        nrDeltaT (float): time difference between the peak of the (2,2) mode and the reference time
        fits_dict (dict): dictionary containing the input-value fits

    Returns:
        Dict: dictionary containing the NQC coefficients
    """

    coeffs = {}

    eta = (m1 * m2) / ((m1 + m2) * (m1 + m2))
    # Eq (3) in LIGO DCC document T1100433v2
    rOmega = r * omega_orb
    q1 = pr**2 / rOmega**2
    q2 = q1 / r
    q3 = q2 / np.sqrt(r)

    # Eq (4) in LIGO DCC document T1100433v2
    p1 = pr / rOmega
    p2 = p1 * pr**2

    # See below Eq(9)
    q1LM = q1 * amplitude
    q2LM = q2 * amplitude
    q3LM = q3 * amplitude

    # Compute the attachment time
    # This is given by Eq.(41) of https://dcc.ligo.org/DocDB/0186/T2300060/001/SEOBNRv5HM.pdf
    # with time_peak <-> t_ISCO and nrDeltaT <-> - DeltaT_22
    # i.e. nrDeltaT > 0 means the attachment time is before t_ISCO
    nrTimePeak = time_peak - nrDeltaT
    if nrTimePeak > time[-1]:
        # If the predicted time is after the end of the dynamics
        # use the end of the dynamics
        nrTimePeak = time[-1]
        if ell == 5 and m == 5:
            nrTimePeak -= 10  # (5,5) mode always 10 M before peak of (2,2)

    # Compute amplitude NQCs (a1,a2,a3) by solving Eq.(10) of T1100433v2
    # Evaluate the Qs at the right time and build Q matrix (LHS)
    idx = np.argmin(np.abs(time - nrTimePeak))
    N = 5
    left = np.max((0, idx - N))
    right = np.min((idx + N, len(time)))

    intrp_q1LM = InterpolatedUnivariateSpline(time[left:right], q1LM[left:right])
    intrp_q2LM = InterpolatedUnivariateSpline(time[left:right], q2LM[left:right])
    intrp_q3LM = InterpolatedUnivariateSpline(time[left:right], q3LM[left:right])

    Q = np.zeros((3, 3))

    Q[:, 0] = intrp_q1LM.derivatives(nrTimePeak)[:-1]
    Q[:, 1] = intrp_q2LM.derivatives(nrTimePeak)[:-1]
    Q[:, 2] = intrp_q3LM.derivatives(nrTimePeak)[:-1]

    # Build the RHS of Eq.(10)
    # Compute the NR fits
    nra = eta * fits_dict["amp"][(ell, m)]
    nraDot = eta * fits_dict["damp"][(ell, m)]
    nraDDot = eta * fits_dict["ddamp"][(ell, m)]

    # Compute amplitude and derivatives at the right time
    intrp_amp = InterpolatedUnivariateSpline(time[left:right], amplitude[left:right])

    amp, damp, ddamp = intrp_amp.derivatives(nrTimePeak)[:-1]

    # Assemble RHS
    amps = np.array([nra - amp, nraDot - damp, nraDDot - ddamp])

    # Solve the equation Q*coeffs = amps
    res = np.linalg.solve(Q, amps)

    coeffs["a1"] = res[0]
    coeffs["a2"] = res[1]
    coeffs["a3"] = res[2]

    # Now we (should) have calculated the a values. 
    # We now compute the frequency NQCs (b1,b2) by solving Eq.(11) of T1100433v2
    # Populate the P matrix in LHS of Eq.(11) of T1100433v2
    intrp_p1 = InterpolatedUnivariateSpline(time[left:right], p1[left:right])
    intrp_p2 = InterpolatedUnivariateSpline(time[left:right], p2[left:right])
    P = np.zeros((2, 2))

    P[:, 0] = -intrp_p1.derivatives(nrTimePeak)[1:-1]
    P[:, 1] = -intrp_p2.derivatives(nrTimePeak)[1:-1]

    # Build the RHS of Eq.(11)
    # Compute frequency and derivative at the right time
    intrp_phase = InterpolatedUnivariateSpline(time[left:right], phase[left:right])
    omega, omegaDot = intrp_phase.derivatives(nrTimePeak)[1:-1]

    # Since the phase can be decreasing, we need to take care not to have a -ve frequency
    if omega * omegaDot > 0.0:
        omega = np.abs(omega)
        omegaDot = np.abs(omegaDot)
    else:
        omega = np.abs(omega)
        omegaDot = -np.abs(omegaDot)

    # Compute the NR fits
    nromega = np.abs(fits_dict["omega"][(ell, m)])
    nromegaDot = np.abs(fits_dict["domega"][(ell, m)])

    # Assemble RHS
    omegas = np.array([nromega - omega, nromegaDot - omegaDot])

    # Solve the equation P*coeffs = omegas
    res = np.linalg.solve(P, omegas)
    coeffs["b1"] = res[0]
    coeffs["b2"] = res[1]
    return coeffs


def EOBNonQCCorrection(r, phi, pr, pphi, omega, coeffs):
    """
    Evaluate the NQC correction, given the coefficients.

    Args:
        r (numpy.ndarray): r along the dynamics
        phi (numpy.ndarray): phase along the dynamics
        pr (numpy.ndarray): pr along the dynamics
        omega (numpy.ndarray): omega along the dynamics
        coeffs (dict): dictionary containing the NQC coefficients

    Returns:
        numpy.ndarray: the NQC corrections along the dynamics
        
    """
    sqrtR = np.sqrt(r)
    rOmega = r * omega
    rOmegaSq = rOmega * rOmega
    p = pr
    mag = 1.0 + (p * p / rOmegaSq) * (
        coeffs["a1"]
        + coeffs["a2"] / r
        + (coeffs["a3"] + coeffs["a3S"]) / (r * sqrtR)
        + coeffs["a4"] / (r * r)
        + coeffs["a5"] / (r * r * sqrtR)
    )
    phase = coeffs["b1"] * p / rOmega + p * p * p / rOmega * (
        coeffs["b2"] + coeffs["b3"] / sqrtR + coeffs["b4"] / r
    )

    nqc = mag * np.cos(phase) + 1j * mag * np.sin(phase)
    return nqc



"""
Contains functions associated with evolving the equations of motion.
"""

from typing import Callable

import numpy as np
import pygsl_lite.errno as errno
import pygsl_lite.odeiv2 as odeiv2
from numba import *
from numba import jit, types

from ..fits.fits_Hamiltonian import dSO as dSO_poly_fit
from ..hamiltonian import Hamiltonian
from ..utils.math_ops_opt import *
from .initial_conditions_aligned_precessing import computeIC_augm
from .pn_evolution_opt import (
    rhs_wrapper,
    compute_omega_orb,
    compute_quasiprecessing_PNdynamics_opt,
    build_splines_PN,
)
from .integrate_ode import interpolate_dynamics, iterative_refinement
from ..utils.containers import EOBParams
from ..utils.math_ops_opt import my_norm

# Test cythonization of PN equations

from scipy.interpolate import CubicSpline
from scipy.signal import argrelmin


step = odeiv2.pygsl_lite_odeiv2_step
_control = odeiv2.pygsl_lite_odeiv2_control
evolve = odeiv2.pygsl_lite_odeiv2_evolve


class control_y_new(_control):
    def __init__(self, eps_abs, eps_rel):
        a_y = 1
        a_dydt = 1
        _control.__init__(self, eps_abs, eps_rel, a_y, a_dydt, None)


# Function to terminate the EOB evolution
def check_terminal(
    r: float,
    omega: float,
    drdt: float,
    dprdt: float,
    omega_circ: float,
    omega_previous: float,
    r_previous: float,
    omegaPN_f: float,
) -> int:
    """
    Check termination condition of the EOB evolution.

    Args:
        r (float): Orbital separation
        omega (float): Orbital frequency
        drdt (float): Time derivative of r
        dprdt (float): Time derivative of prstar
        omega_circ (float): Circular orbital frequency
        omega_previous (float): Orbital frequency at the previous timestep
        r_previous (float): Orbital separation at the previous timestep
        omegaPN_f (float): Final orbital frequency reached in the spin-precessing PN evolution

    Returns:
        (int): If >0 terminates EOB dynamics
    """

    if r <= 1.4:
        # print(f"r = {r} < 1.4, omega = {omega}")
        return 1

    if np.isnan(omega) or np.isnan(drdt) or np.isnan(dprdt) or np.isnan(omega_circ):
        # print(f"nan omega: r = {r}, omega = {omega}")
        return 2

    if omega < omega_previous:
        # print(f"r = {r}, omega = {omega}, omega_previous = {omega_previous}")
        return 3

    if r < 3:

        if drdt > 0:
            # print(f"drdt>0 : r = {r}, omega = {omega}, omega_previous = {omega_previous}")
            return 5

        if dprdt > 0:
            # print(f"dprdt >0 : r = {r}, omega = {omega}, omega_previous = {omega_previous}")
            return 6

        if omega_circ > 1:
            # print(f"omega_circ >1 : r = {r}, omega = {omega}, omega_previous = {omega_previous}")
            return 7

        if r > r_previous:
            # print(f"r> r_previous : r = {r}, omega = {omega}, omega_previous = {omega_previous}, r_previous = {r_previous}")
            return 8

    if omega > omegaPN_f:
        # print(f"omega> omegaPN_f : r = {r}, omega = {omega}, omegaPN_f = {omegaPN_f}")
        return 9

    return 0


def compute_dynamics_prec_opt(
    omega_ref: float,
    omega_start: float,
    omegaPN_f: float,
    H: Hamiltonian,
    RR: Callable,
    m_1: float,
    m_2: float,
    splines: dict,
    t_pn: np.array,
    dynamics_pn: np.array,
    params: EOBParams,
    rtol: float = 1e-12,
    atol: float = 1e-12,
    step_back: float = 250.0,
    y_init=None,
    initial_conditions: str = "adiabatic",
    initial_conditions_postadiabatic_type: str = "analytic",
):
    """
    Function to perform a non-precessing EOB evolution with the spins modified
    at every timestep according to the values from the precessing-spin PN
    evolution.

    Args:
        omega_ref (float): Reference orbital frequency at which the spins are defined
        omega_start (float): Starting orbital frequency
        omegaPN_f (float): Final orbital frequency from the precessing-spin PN evolution
        H (Hamiltonian): Hamiltonian class
        RR (Callable): RR force
        m_1 (float): Mass component of the primary
        m_2 (float): Mass component of the secondary
        splines (dict): Dictionary containing the splines in orbital frequency of the vector components of the spins, LN and L as
                        well as the spin projections onto LN and L
        t_pn (np.array): Time array of the PN evolution of the spins and Newtonian angular momentum.
        dynamics_pn (np.array): Array of the spin-precessing PN evolution. It contains the Newtonian angular momentum, the dimensionful spin vectors and the PN orbital frequency.
        params (EOBParams): Container of additional inputs
        rtol (float, optional): Relative tolerance for EOB integration. Defaults to 1e-12
        atol (float, optional): Absolute tolerance for EOB integration. Defaults to 1e-12
        step_back (float, optional): Amount of time to step back for fine interpolation. Defaults to 250.
        y_init (np.ndarray, optional): Initial condition vector (r,phi,pr,pphi).
        initial_conditions (str, optional): Type of initial conditions for the ODE evolution ('adiabatic' or 'postadiabatic').
        initial_conditions_postadiabatic_type (str, optional): Type of postadiabatic initial conditions for the ODE evolution ('analytic' or 'numeric').

    Returns:
        (tuple): Low and high sampling rate dynamics, unit Newtonian orbital angular momentum, assembled dynamics
                 and the index splitting the low and high sampling rate dynamics
    """

    # Step 3 : Evolve EOB dynamics

    # Step 3.1) Update spin parameters and calibration coeffs (only dSO) at the start of EOB integration

    X1 = params.p_params.X_1
    X2 = params.p_params.X_2

    tmp = splines["everything"](omega_start)
    chi1_LN_start = tmp[0]
    chi2_LN_start = tmp[1]
    chi1_L_start = tmp[2]
    chi2_L_start = tmp[3]
    chi1_v_start = tmp[4:7]
    chi2_v_start = tmp[7:10]

    params.p_params.omega = omega_start
    params.p_params.chi1_v[:] = chi1_v_start
    params.p_params.chi2_v[:] = chi2_v_start

    params.p_params.chi_1, params.p_params.chi_2 = chi1_LN_start, chi2_LN_start
    params.p_params.chi1_L, params.p_params.chi2_L = chi1_L_start, chi2_L_start

    params.p_params.update_spins(chi1_LN_start, chi2_LN_start)

    ap_start = chi1_LN_start * X1 + chi2_LN_start * X2
    am_start = chi1_LN_start * X1 - chi2_LN_start * X2

    dSO_start = dSO_poly_fit(params.p_params.nu, ap_start, am_start)
    H.calibration_coeffs["dSO"] = dSO_start

    # Step 3.2: compute the initial conditions - uses the aligned-spin ID
    if y_init is None:
        if initial_conditions == "adiabatic":
            r0, pphi0, pr0 = computeIC_augm(
                omega_start,
                H,
                RR,
                chi1_v_start,
                chi2_v_start,
                m_1,
                m_2,
                params=params,
            )

        elif initial_conditions == "postadiabatic":
            from .initial_conditions_precessing_postadiabatic import compute_IC_PA

            r0, pphi0, pr0 = compute_IC_PA(
                omega_ref,
                omega_start,
                H,
                RR,
                chi1_v_start,
                chi2_v_start,
                m_1,
                m_2,
                splines,
                t_pn,
                dynamics_pn,
                params=params,
                postadiabatic_type=initial_conditions_postadiabatic_type,
            )
        y0 = np.array([r0, 0.0, pr0, pphi0])

    else:

        y0 = y_init.copy()
        r0 = y0[0]

    # Step 3.3: now integrate the dynamics
    sys = odeiv2.system(rhs_wrapper, None, 4, [H, RR, m_1, m_2, params])

    T = odeiv2.step_rkf45
    s = step(T, 4)
    c = control_y_new(atol, rtol)
    e = evolve(4)

    t = 0
    t1 = 1.0e9

    y = y0

    if y_init is None:
        h = 2 * np.pi / omega_start / 100.0
    else:
        h = 0.1

    # Use an agnostic and small initial step for the integrator
    h = 0.1
    res_gsl = []
    ts = []
    omegas = []
    augm_dyn = []

    # Convert to numpy array to get the correct value
    augm_dyn.append(
        [
            params.p_params.H_val,
            params.p_params.omega,
            params.p_params.omega_circ,
            params.p_params.chi_1,
            params.p_params.chi_2,
        ]
    )
    ts.append(0.0)
    res_gsl.append(y)
    omegas.append(params.p_params.omega)

    omega_previous = omega_start
    r_previous = r0

    X1 = params.p_params.X_1
    X2 = params.p_params.X_2

    peak_omega = False
    peak_pr = False

    while t < t1:

        # Take a step
        status, t, h, y = e.apply(c, s, sys, t, t1, h, y)
        if status != errno.GSL_SUCCESS:
            print("break status", status)
            break
        # Compute the error for the step controller
        e.get_yerr()

        r = y[0]

        if np.isnan(r):
            # print(f"t = {t}, r = {r}, h = {h}")
            break
        else:
            # Append the last step
            res_gsl.append(y)
            ts.append(t)
            augm_dyn.append(
                [
                    params.p_params.H_val,
                    params.p_params.omega,
                    params.p_params.omega_circ,
                    params.p_params.chi_1,
                    params.p_params.chi_2,
                ]
            )

        # Handle termination conditions
        if r <= 6:

            deriv = rhs_wrapper(t, y, [H, RR, m_1, m_2, params])
            drdt = deriv[0]
            omega = deriv[1]
            omega_previous = omegas[-1]
            if np.isnan(omega):
                res_gsl = res_gsl[:-1]
                ts = ts[:-1]
                augm_dyn = augm_dyn[:-1]
                break

            else:

                omegas.append(omega)
                dprdt = deriv[2]
                omega_circ = params.p_params.omega_circ

            # check termination conditions
            termination = check_terminal(
                r, omega, drdt, dprdt, omega_circ, omega_previous, r_previous, omegaPN_f
            )
            if termination:
                if termination == 3:
                    peak_omega = True
                if termination == 6:
                    peak_pr = True
                break

            omega_previous = omega

        else:
            omega = compute_omega_orb(t, y, H, RR, m_1, m_2, params)
            omegas.append(omega)

        # Update spin parameters and calibration coeffs (only dSO)
        params.p_params.omega = omega
        omega_circ = params.p_params.omega_circ

        # Update spins according to the new value of orbital frequency
        tmp = splines["everything"](omega)
        chi1_LN = tmp[0]
        chi2_LN = tmp[1]
        chi1_L = tmp[2]
        chi2_L = tmp[3]
        chi1_v = tmp[4:7]
        chi2_v = tmp[7:10]
        tmp_LN = tmp[10:13]

        params.p_params.chi1_v[:] = chi1_v
        params.p_params.chi2_v[:] = chi2_v
        # params.p_params.lN[:] = tmp_LN#/my_norm(tmp_LN)

        ap = chi1_LN * X1 + chi2_LN * X2
        am = chi1_LN * X1 - chi2_LN * X2

        # Check for possible extrapolation in the spin values
        if abs(ap) > 1 or abs(am) > 1:
            # print(f"r = {r}, omega = {omega}, ap = {ap}, am = {am}")
            break

        dSO_new = dSO_poly_fit(params.p_params.nu, ap, am)
        H.calibration_coeffs["dSO"] = dSO_new

        params.p_params.chi_1 = chi1_LN
        params.p_params.chi_2 = chi2_LN

        params.p_params.chi1_L = chi1_L
        params.p_params.chi2_L = chi2_L

        # Remember to update the derived spin quantities which enter the flux
        params.p_params.update_spins(params.p_params.chi_1, params.p_params.chi_2)

    # Step 3.4: After the integration assemble quantities and apply a roll-off of the timestep
    #           to avoid abrupt changes in the splines
    ts = np.array(ts)
    dyn = np.array(res_gsl)
    omega_eob = np.array(omegas)
    augm_dyn = np.array(augm_dyn)

    dyn = np.c_[dyn, augm_dyn]

    # Step 3.5: Iterative procedure to find the peak of omega or pr to have a more robust model
    #           under perturbations

    ##################################################################

    # Estimate of the starting point of fine integration
    # Same refinement as in the aligned-spin model
    t_stop = ts[-1]
    if peak_omega:
        t_desired = t_stop - step_back - 50
    else:
        t_desired = t_stop - step_back

    idx_close = np.argmin(np.abs(ts - t_desired))
    t_fine = ts[idx_close:]
    dyn_fine = np.c_[t_fine, dyn[idx_close:]]
    omega_fine = omega_eob[idx_close:]

    t_peak = None
    if peak_omega:
        idx_max = np.argmax(omega_fine)
        omega_peak = omega_fine[idx_max]
        t_peak = t_fine[idx_max]

        if idx_max != len(omega_fine):
            omega_idxm1 = omega_fine[idx_max - 1]
            t_idxm1 = t_fine[idx_max - 1]

            omega_idx0 = omega_peak
            t_idx0 = t_fine[idx_max]

            omega_idx1 = omega_fine[idx_max + 1]
            t_idx1 = t_fine[idx_max + 1]

            t_peak, omega_peak = parabolaExtrema(
                omega_idx0, omega_idxm1, omega_idx1, t_idx0, t_idxm1, t_idx1
            )

    if peak_pr:
        intrp = CubicSpline(t_fine, dyn_fine[:, 3])
        left = t_fine[-1] - 10
        right = t_fine[-1]
        t_peak = iterative_refinement(intrp.derivative(), [left, right], pr=True)

    t_roll, dyn_roll, omega_roll = ts, dyn, omega_eob
    res = np.c_[t_roll, dyn_roll, omega_roll]

    if peak_omega or peak_pr:
        t_start = max(t_peak - step_back, dyn_fine[0, 0])
        idx_fine_start = np.argmin(np.abs(t_roll - t_start))

        idx_close = idx_fine_start

        t_fine = t_roll[idx_close:]
        dyn_fine = np.c_[t_fine, dyn_roll[idx_close:]]
        omega_fine = omega_roll[idx_close:]

    t_roll, dyn_roll, omega_roll, idx_close = transition_dynamics_v2(
        ts, dyn, omega_eob, idx_close
    )

    t_fine = t_roll[idx_close:]
    dyn_fine = np.c_[t_fine, dyn_roll[idx_close:]]
    omega_fine = omega_roll[idx_close:]

    dynamics_low = np.c_[t_roll[:idx_close], dyn_roll[:idx_close]]
    omega_low = omega_roll[:idx_close]

    # Add LN to the array so that it is also interpolated onto the fine sampling rate dynamics
    dyn_fine = interpolate_dynamics(dyn_fine, peak_omega=t_peak, step_back=step_back)

    # Remove points where omega starts decreasing as we need it
    # to be monotonically increasing as it is used for interpolation

    omega_dyn = dyn_fine[:, 6]
    om_diff = np.diff(omega_dyn)
    idx_omdiff = np.where(om_diff < 0)[0]

    if idx_omdiff.size != 0:
        idx_final = idx_omdiff[0]
        dyn_fine = dyn_fine[: idx_final + 1]

    # Define the dynamics
    dynamics_fine = dyn_fine

    # Full dynamics array
    dynamics = np.vstack((dynamics_low, dynamics_fine))

    # Return EOB dynamics, LN vectors,  PN stuff
    return (dynamics_low, dynamics_fine, dynamics, idx_close)


def transition_dynamics_v2(
    ts: np.ndarray, dyn: np.ndarray, omega_eob: np.ndarray, idx_restart: int
):
    """
    Function to transition from a point dyn1,t1final to a point (dyn2,tfinal2)

    Args:

         ts (np.ndarray): Time array
         dyn (np.ndarray): Dynamics array  (r,phi,pr,pphi)
         omega_eob (np.ndarray): Orbital frequency array
         idx_restart (int): Index which separates the low sampling rate dynamics and the high sampling rate dynamics.

    Returns:
         (tuple): Time array, dynamics, orbital frequency and index splitting the low and high sampling rate dynamics
    """

    # If the closest point within step back is actually the last point, step back more
    dyn_low = dyn[:idx_restart, :]
    dyn_fine = dyn[idx_restart:, :]

    t_low = ts[:idx_restart]
    t_fine = ts[idx_restart:]

    t_low_last = t_low[-1]
    t_fine_init = t_fine[0]

    if t_fine_init - t_low_last > 2:

        dt_low_last = t_low[-1] - t_low[-2]
        dt_fine = t_fine[1] - t_fine[0]
        dt_fine = 0.1

        t = t_fine[0] - dt_fine
        t_new = []

        step_multiplier = 1.3
        dt = dt_fine

        while True:
            t_new.append(t)

            if step_multiplier * dt < dt_low_last:
                dt *= step_multiplier

            t -= dt

            if t < t_low_last:
                break

        t_new = t_new[::-1]

        window = 50
        while idx_restart < window:
            window -= 5

        t_middle = ts[idx_restart - window : idx_restart + window]
        dyn_window = dyn[idx_restart - window : idx_restart + window]

        dyn_interp = CubicSpline(t_middle, dyn_window[:, :])
        omega_interp = CubicSpline(
            t_middle, omega_eob[idx_restart - window : idx_restart + window]
        )

        dyn_middle = dyn_interp(t_new)
        omega_middle = omega_interp(t_new)

        # Separate low and high SR omega
        omega_low = omega_eob[:idx_restart]
        omega_fine = omega_eob[idx_restart:]

        time = np.concatenate((t_low, t_new[:-1], t_fine))
        dynamics = np.vstack((dyn_low, dyn_middle[:-1, :], dyn_fine))
        omega = np.concatenate((omega_low, omega_middle[:-1], omega_fine))

        idx_restart_v1 = (
            len(omega_low) + len(omega_middle) - 1
        )  # idx_restart -1  + window_length
    else:

        dynamics = dyn
        time = ts
        omega = omega_eob
        idx_restart_v1 = idx_restart

    return time, dynamics, omega, idx_restart_v1


def parabolaExtrema(
    ff0: float, ffm1: float, ff1: float, tt0: float, ttm1: float, tt1: float
):
    """
    Compute the extremum of a parabola from 3 points (idx-1, idx, idx+1).

    Args:
        ff0 (float): value of the function at the index idx
        ffm1 (float): value of the function at the index idx-1
        ff1 (float): value of the function at the index idx+1
        tt0 (float): value of the time array at the index idx
        ttm1 (float): value of the time array at the index idx-1
        tt1 (float): value of the time array at the index idx+1

    Returns:
        tuple: time of the extremum, and value of the function at the extremum
    """

    aa = (ffm1 * (tt0 - tt1) + ff0 * (tt1 - ttm1) + ff1 * (-tt0 + ttm1)) / (
        (tt0 - tt1) * (tt0 - ttm1) * (tt1 - ttm1)
    )
    bb = (
        ffm1 * (-tt0 * tt0 + tt1 * tt1)
        + ff1 * (tt0 * tt0 - ttm1 * ttm1)
        + ff0 * (-tt1 * tt1 + ttm1 * ttm1)
    ) / ((tt0 - tt1) * (tt0 - ttm1) * (tt1 - ttm1))
    cc = (
        ffm1 * tt0 * (tt0 - tt1) * tt1
        + ff0 * tt1 * (tt1 - ttm1) * ttm1
        + ff1 * tt0 * ttm1 * (-tt0 + ttm1)
    ) / ((tt0 - tt1) * (tt0 - ttm1) * (tt1 - ttm1))

    textremum = -bb / (2.0 * aa)
    ff_extremum = aa * textremum * textremum + bb * textremum + cc

    return textremum, ff_extremum


def compute_dynamics_quasiprecessing(
    omega_ref: float,
    omega_start: float,
    H: Hamiltonian,
    RR: Callable,
    m_1: float,
    m_2: float,
    chi_1: np.ndarray,
    chi_2: np.ndarray,
    params: EOBParams,
    rtol: float = 1e-12,
    atol: float = 1e-12,
    step_back: float = 250,
    y_init=None,
    initial_conditions=None,
    initial_conditions_postadiabatic_type=None,
):
    """
    Compute the dynamics starting from omega_start, with spins
    defined at omega_ref.

    First, PN evolution equations are integrated (including backwards in time)
    to get spin and orbital angular momentum. From that we construct splines
    either in time or orbital frequency for the PN quantities. Given the splines
    we now integrate aligned-spin EOB dynamics where at every step the projections
    of the spins onto orbital angular momentum is computed via the splines.

    Args:
        omega_ref (float): Reference frequency
        omega_start (float): Starting frequency
        H (Hamiltonian): Hamiltonian to use
        RR (Callable): RR force to use
        m_1 (float): Mass of primary
        m_2 (float): Mass of secondary
        chi_1 (np.ndarray): Dimensionless spin of the primary
        chi_2 (np.ndarray): Dimensionless spin of the secondary
        params (EOBParams): Container of additional inputs
        rtol (float, optional): Relative tolerance for EOB integration. Defaults to 1e-12.
        atol (float, optional): Absolute tolerance for EOB integration. Defaults to 1e-12.
        step_back (float, optional): Amount of time to step back for fine interpolation. Defaults to 250.
        y_init (np.ndarray, optional): Initial condition vector (r,phi,pr,pphi).
        initial_conditions (str, optional): Type of initial conditions for the ODE evolution ('adiabatic' or 'postadiabatic').
        initial_conditions_postadiabatic_type (str, optional): Type of postadiabatic initial conditions for the ODE evolution ('analytic' or 'numeric').

    Returns:
        tuple: Aligned-spin EOB dynamics, PN time, PN dynamics, PN splines
    """

    # Step 1: Compute PN dynamics

    if initial_conditions == "postadiabatic":

        combined_t, combined_y = compute_quasiprecessing_PNdynamics_opt(
            omega_ref, 0.9 * omega_start, m_1, m_2, chi_1, chi_2
        )
    else:

        combined_t, combined_y = compute_quasiprecessing_PNdynamics_opt(
            omega_ref, omega_start, m_1, m_2, chi_1, chi_2
        )
    # Compute last value of the orbital frequency from the PN evolution (to be used for
    # the termination conditions of the non-precessing evolution)
    omegaPN_f = combined_y[:, -1][-1]

    # Step 2: Interpolate PN dynamics
    splines = build_splines_PN(combined_t, combined_y, m_1, m_2, omega_start)

    # Step 3: Evolve the non-precessing EOB evolution equations
    (dynamics_low, dynamics_fine, dynamics, idx_restart) = compute_dynamics_prec_opt(
        omega_ref,
        omega_start,
        omegaPN_f,
        H,
        RR,
        m_1,
        m_2,
        splines,
        combined_t,
        combined_y,
        params,
        rtol=rtol,
        atol=atol,
        step_back=step_back,
        y_init=y_init,
        initial_conditions=initial_conditions,
        initial_conditions_postadiabatic_type=initial_conditions_postadiabatic_type,
    )

    return (
        dynamics_low,
        dynamics_fine,
        combined_t,
        combined_y,
        splines,
        dynamics,
        idx_restart,
    )



# cython: language_level=3

"""
Contains the functions needed for computing the post-adiabatic dynamics as well as the combined dynamics
(post-adiabatic + final part of the inspiral evolved using the usual EOB dynamics)
"""

cimport cython
from typing import Dict
import numpy as np
from libc.math cimport log, sqrt, exp, abs,fabs, tgamma,sin,cos, tanh, sinh, asinh

from .initial_conditions_aligned_opt import computeIC_opt as computeIC


from pygsl_lite import  roots, errno
from scipy import optimize
from scipy.interpolate import CubicSpline, InterpolatedUnivariateSpline
from scipy import integrate

from lalinference.imrtgr import nrutils

from .integrate_ode import compute_dynamics_opt as compute_dynamics
from .integrate_ode import augment_dynamics
from pyseobnr.eob.utils.containers cimport EOBParams
from pyseobnr.eob.hamiltonian.Hamiltonian_C cimport Hamiltonian_C
from pyseobnr.eob.waveform.waveform cimport RadiationReactionForce

fin_diff_coeffs_order_9 = np.array([
    [-761./280., 8., -14., 56./3., -35./2., 56./5., -14./3., 8./7., -1./8.],
    [-1./8., -223./140., 7./2., -7./2., 35./12., -7./4., 7./10., -1./6., 1./56.],
    [1./56., -2./7., -19./20., 2., -5./4., 2./3., -1./4., 2./35., -1./168.],
    [-1./168., 1./14., -1./2., -9./20., 5./4., -1./2., 1./6., -1./28., 1./280.],
    [1./280., -4./105., 1./5., -4./5., 0, 4./5., -1./5., 4./105., -1./280.],
    [-1./280., 1./28., -1./6., 1./2., -5./4., 9./20., 1./2., -1./14., 1./168.],
    [1./168., -2./35., 1./4., -2./3., 5./4., -2., 19./20., 2./7., -1./56.],
    [-1./56., 1./6., -7./10., 7./4., -35./12., 7./2., -7./2., 223./140., 1./8.],
    [1./8., -8./7., 14./3., -56./5., 35./2., -56./3., 14., -8., 761./280.],
])

interpolated_integral_order_3 = [
    [3./8., 19./24., -5./24., 1./24.],
    [-1./24., 13./24., 13./24., -1./24.],
    [1./24., -5./24., 19./24., 3./8.],
]

interpolated_integral_order_5 = [
    [95./288., 1427./1440., -133./240., 241./720., -173./1440., 3./160.],
    [-3./160., 637./1440., 511./720., -43./240., 77./1440., -11./1440.],
    [11./1440., -31./480., 401./720., 401./720., -31./480., 11./1440.],
    [-11./1440., 77./1440., -43./240., 511./720., 637./1440., -3./160.],
    [3./160., -173./1440., 241./720., -133./240., 1427./1440., 95./288.],
]

interpolated_integral_order_7 = [
    [5257./17280, 139849./120960, -(4511./4480), 123133./120960, -(88547./120960), 1537./4480, -(11351./120960), 275./24192],
    [-(275./24192), 5311./13440, 11261./13440, -(44797./120960), 2987./13440, -(1283./13440), 2999./120960, -(13./4480)],
    [13./4480, -(4183./120960), 6403./13440, 9077./13440, -(20227./120960), 803./13440, -(191./13440), 191./120960],
    [-(191./120960), 1879./120960, -(353./4480), 68323./120960, 68323./120960, -(353./4480), 1879./120960, -(191./120960)],
    [191./120960, -(191./13440), 803./13440, -(20227./120960), 9077./13440, 6403./13440, -(4183./120960), 13./4480],
    [-(13./4480), 2999./120960, -(1283./13440), 2987./13440, -(44797./120960), 11261./13440, 5311./13440, -(275./24192)],
    [275./24192, -(11351./120960), 1537./4480, -(88547./120960), 123133./120960, -(4511./4480), 139849./120960, 5257./17280],
]


@cython.profile(True)
@cython.linetrace(True)
@cython.cdivision(True)
@cython.boundscheck(False)
cdef double single_deriv(double[:] y,double h,double[:] coeffs):
    cdef int i
    cdef double total = 0.0
    for i in range(9):
        total+=coeffs[i]*y[i]
    total/=h
    return total

@cython.profile(True)
@cython.linetrace(True)
@cython.cdivision(True)
cpdef fin_diff_derivative(
    x,
    y,
    n: int = 8,
):
    """
    Compute 8th order finite difference derivative,
    assuming an equally spaced grid. Correctly uses
    asymmetric stencils at the end points.
    """

    dy_dx = np.zeros(x.size)
    cdef double h = fabs(x[1] - x[0])
    cdef int size = x.shape[0]
    cdef int i
    for i in range(size):
        if i == 0:
            dy_dx[i] = single_deriv(y[0:9],h,fin_diff_coeffs_order_9[0])
        elif i == 1:
            dy_dx[i] = single_deriv(y[0:9],h,fin_diff_coeffs_order_9[1])
        elif i == 2:
            dy_dx[i] = single_deriv(y[0:9],h,fin_diff_coeffs_order_9[2])
        elif i == 3:
            dy_dx[i] = single_deriv(y[0:9],h,fin_diff_coeffs_order_9[3])
        elif i == size - 4:
            dy_dx[i] = single_deriv(y[-9:],h,fin_diff_coeffs_order_9[5])
        elif i == size - 3:
            dy_dx[i] = single_deriv(y[-9:],h,fin_diff_coeffs_order_9[6])
        elif i == size - 2:
            dy_dx[i] = single_deriv(y[-9:],h,fin_diff_coeffs_order_9[7])
        elif i == size - 1:
            dy_dx[i] = single_deriv(y[-9:],h,fin_diff_coeffs_order_9[8])
        else:
            dy_dx[i] = single_deriv(y[i-4:i+5],h,fin_diff_coeffs_order_9[4])

    return dy_dx

@cython.profile(True)
@cython.linetrace(True)
cpdef Kerr_ISCO(
    double chi1,
    double chi2,
    double m1,
    double m2,
):
    """
    Compute the Kerr ISCO radius and angular momentum
    from the remnant spin predicted by NR fits
    """
    a = nrutils.bbh_final_spin_non_precessing_HBR2016(
            m1, m2, chi1, chi2, version="M3J4"
    )
    # Compute the ISCO radius for this spin
    Z_1 = 1+(1-a**2)**(1./3)*((1+a)**(1./3)+(1-a)**(1./3))
    Z_2 = np.sqrt(3*a**2+Z_1**2)
    r_ISCO = 3+Z_2-np.sign(a)*np.sqrt((3-Z_1)*(3+Z_1+2*Z_2))

    # Compute the ISCO L for this spin
    L_ISCO = 2/(3*np.sqrt(3))*(1+2*np.sqrt(3*r_ISCO)-2)
    return np.array([r_ISCO,L_ISCO])

@cython.profile(True)
@cython.linetrace(True)
def Newtonian_j0(r):
    """
    Newtonian expression for orbital angular momentum using
    consistent normalization.
    """
    return np.sqrt(r)

@cython.profile(True)
@cython.linetrace(True)
cpdef double j0_eqn(double j0_sol, double r, Hamiltonian_C H, double chi_1, double chi_2, double m_1, double m_2,
            double[:] q, double[:] p):
    q[0] = r
    p[1]= j0_sol

    cdef (double,double,double,double) dH_dq = H.grad(q, p, chi_1, chi_2, m_1, m_2)
    cdef double dH_dr = dH_dq[0]
    return dH_dr

cpdef compute_adiabatic_solution(
    double[:] r,
    Hamiltonian_C H,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    double[::1] q,
    double[::1] p,
    double tol=1e-12,
):
    """
    Compute the adiabatic solution for the orbital angular momentum.
    Corresponds to Eq.(2.5) of arXiv:2105.06983
    """
    cdef int i
    cdef double[:] j0 = Newtonian_j0(r)

    for i in range(r.shape[0]):
        j0_solution = optimize.root(
            j0_eqn,
            j0[i],
            args=(r[i], H, chi_1, chi_2, m_1, m_2,q,p),
            tol=tol,
        )
        j0[i] = j0_solution.x

    return j0

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.initializedcheck(False)
@cython.profile(True)
@cython.linetrace(True)
cpdef double pr_eqn(
    double pr_sol,
    double r,
    double pphi,
    double dpphi_dr,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    EOBParams params,
    double[::1] q,
    double[::1] p
):
    """
    Evaluate the equation for odd-PA orders, corresponding to corrections to pr.
    See Eq. (2.6) of arXiv:2105.06983
    """
    q[0] = r
    p[0] = pr_sol
    p[1] = pphi
    cdef double dH_dpr,H_val,csi,omega,omega_circ,result
    cdef double dynamics[6]
    dynamics[:] = H.dynamics(q, p, chi_1, chi_2, m_1, m_2)
    #cdef (double,double,double,double)  dH_dx = dynamics[:4]

    dH_dpr = dynamics[2]
    H_val = dynamics[4]
    csi = dynamics[5]


    params.dynamics.p_circ[1] = pphi
    omega_circ = H.omega(q,params.dynamics.p_circ,chi_1,chi_2,m_1,m_2)

    omega = dynamics[3]

    cdef (double,double) flux = RR.RR(q, p, omega,omega_circ,H_val,params)


    result = dpphi_dr * dH_dpr * csi - flux[1]


    return result

cpdef double pr_eqn_wrapper(pr_sol,args):
    return pr_eqn(pr_sol,*args)

@cython.profile(True)
@cython.linetrace(True)
cpdef compute_pr(
    double[:] r,
    double[:] pr,
    double[:] pphi,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    double[::1] q,
    double[::1] p,
    double tol=1e-12,
    EOBParams params=None,
):
    """
    Compute the value to pr at odd PA orders.
    This is done by numericall solving Eq(2.6) of
    arXiv:2105.06983 at every radial grid point
    """

    cdef double[:] dpphi_dr = - fin_diff_derivative(r, pphi)


    cdef int i
    for i in range(r.shape[0]):
        if np.abs(pr[i])<1e-14:
            x0 = 0.0
            x1 = -3e-2
        else:
            x0 = pr[i]*0.95
            x1 = pr[i]*1.05

        mysys = roots.gsl_function(pr_eqn_wrapper, (
                r[i], pphi[i], dpphi_dr[i],
                H, RR, chi_1, chi_2, m_1, m_2,
                params,q,p
            ))
        solver = roots.brent(mysys)
        solver.set(x1,x0)
        for iter in range(100):
            status = solver.iterate()
            x_lo = solver.x_lower()
            x_up = solver.x_upper()
            status = roots.test_interval(x_lo, x_up, tol,tol)
            result = solver.root()
            if status == errno.GSL_SUCCESS:
                break
        pr[i] = result
    return pr

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.initializedcheck(False)
@cython.profile(True)
@cython.linetrace(True)
cpdef double pphi_eqn(
    double pphi_sol,
    double r,
    double pr,
    double dpr_dr,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    EOBParams params,
    double[::1] q,
    double[::1] p
):
    """
    Evaluate the equation for even-PA orders, corresponding to corrections to pphi.
    See Eq. (2.7) of arXiv:2105.06983
    """
    cdef double dH_dr,dH_dpr,H_val,csi,omega,omega_circ,result
    q[0] = r
    p[0] = pr
    p[1] =  pphi_sol
    cdef double dynamics[6]
    dynamics[:] = H.dynamics(q,p, chi_1, chi_2, m_1, m_2)
    #cdef (double,double,double,double) dH_dx = dynamics[:4]
    dH_dr = dynamics[0]
    dH_dpr = dynamics[2]
    H_val = dynamics[4]
    csi = dynamics[5]

    params.dynamics.p_circ[1] =  pphi_sol
    omega_circ = H.omega(q,params.dynamics.p_circ,chi_1,chi_2,m_1,m_2)

    omega = dynamics[3]


    cdef (double,double) flux = RR.RR(q, p, omega,omega_circ,H_val,params)



    result = dpr_dr * dH_dpr + dH_dr - flux[0] / csi

    return result

cpdef double pphi_eqn_wrapper(pphi_sol,args):
    return pphi_eqn(pphi_sol,*args)

@cython.profile(True)
@cython.linetrace(True)
cpdef compute_pphi(
    double[:] r,
    double[:] pr,
    double[:] pphi,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    double[::1] q,
    double[::1] p,
    double tol=1e-12,
    EOBParams params=None
):
    """
    Compute the correction to pphi at even PA orders.
    This is done by numericall solving Eq(2.7) of
    arXiv:2105.06983 at every radial grid point
    """

    cdef double[:] dpr_dr = - fin_diff_derivative(r, pr)
    cdef int i
    for i in range(r.shape[0]):

        x0 = 0.95*pphi[i]
        x1 = 1.05*pphi[i]
        mysys = roots.gsl_function(pphi_eqn_wrapper, (
                r[i], pr[i], dpr_dr[i], H, RR, chi_1, chi_2, m_1, m_2, params,q,p
            ))
        solver = roots.brent(mysys)
        solver.set(x0,x1)
        for iter in range(100):
            status = solver.iterate()
            x_lo = solver.x_lower()
            x_up = solver.x_upper()
            status = roots.test_interval(x_lo, x_up, tol,tol)
            result = solver.root()
            if status == errno.GSL_SUCCESS:
                break
        pphi[i] = result
    return pphi

@cython.profile(True)
@cython.linetrace(True)
cpdef compute_postadiabatic_solution(
    double[:] r,
    double[:] pphi,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    double[::1] q,
    double[::1] p,
    double tol=1e-12,
    int order=8,
    EOBParams params=None,
):
    """
    Compute the postadiabatic solution iteratively
    """
    pr = np.zeros(r.size)
    cdef int n
    cdef double tol_current
    for n in range(1, order+1):
        tol_current =  1.0e-2/10.0**n
        #print(tol_current)
        parity = n % 2
        if n>=7:
            tol_current=tol
        if parity:
            pr = compute_pr(
                r,
                pr,
                pphi,
                H,
                RR,
                chi_1,
                chi_2,
                m_1,
                m_2,
                q,
                p,
                tol=tol_current,
                params=params,
            )
        else:
            pphi = compute_pphi(
                r,
                pr,
                pphi,
                H,
                RR,
                chi_1,
                chi_2,
                m_1,
                m_2,
                q,
                p,
                tol=tol,
                params=params,
            )

    return pr, pphi

@cython.profile(True)
@cython.linetrace(True)
cpdef compute_postadiabatic_dynamics(
    double omega0,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    double tol=1e-12,
    EOBParams params=None,
    order=8
):
    """Compute the dynamics using PA procedure starting from omega0

    Args:
        omega0 (float): The starting *orbital* frequency in geomtric units
        H (Hamiltonian): The Hamiltonian object to use
        RR (function): The RR function to use
        chi_1 (float): z-component of the primary spin
        chi_2 (float): z-component of the secondary spin
        m_1 (float): mass of the primary
        m_2 (float): mass of the secondary

    Returns:
        np.array: The dynamics results, as (t,q,p)
    """
    cdef double r0
    r0, _, _ = computeIC(
        omega0,
        H,
        RR,
        chi_1,
        chi_2,
        m_1,
        m_2,
        params=params,
    )
    cdef double chi_eff = m_1*chi_1+m_2*chi_2
    cdef double nu = m_1*m_2/(m_1+m_2)**2
    cdef double r_final_prefactor = 2.7+chi_eff*(1-4.*nu)
    r_ISCO, _ = Kerr_ISCO(chi_1, chi_2, m_1, m_2)
    cdef double r_final = max(10.0,r_final_prefactor * r_ISCO)
    cdef double r_switch_prefactor = 1.6
    cdef double r_switch = r_switch_prefactor * r_ISCO

    cdef double dr0 = 0.3
    cdef int r_size = int(np.ceil((r0 - r_final) / dr0))

    if r_size <= 4 or r0<=11.5:
        raise ValueError
    elif r_size < 10:
        r_size = 10

    r, _ = np.linspace(r0, r_final, num=r_size, endpoint=True, retstep=True)
    cdef double[::1] q = np.zeros(2)
    cdef double[::1] p = np.zeros(2)
    pphi = compute_adiabatic_solution(r, H, chi_1, chi_2, m_1, m_2,q,p, tol=tol,)

    pr, pphi = compute_postadiabatic_solution(
        r,
        pphi,
        H,
        RR,
        chi_1,
        chi_2,
        m_1,
        m_2,
        q,
        p,
        tol=tol,
        order=order,
        params=params,
    )
    dt_dr = np.zeros(r.shape[0])
    dphi_dr = np.zeros(r.shape[0])
    cdef int i
    cdef double dH_dpr,dH_dpphi,csi
    cdef double dyn[6]
    for i in range(r.shape[0]):
        q = np.array([r[i], 0])
        p = np.array([pr[i], pphi[i]])

        dyn[:] = H.dynamics(q, p, chi_1, chi_2, m_1, m_2)
        dH_dpr = dyn[2]
        dH_dpphi = dyn[3]


        csi = dyn[5]
        dt_dr[i] = 1 / (dH_dpr * csi)
        dphi_dr[i] = dH_dpphi / (dH_dpr * csi)


    t = cumulative_integral(r, dt_dr)

    phi = cumulative_integral(r, dphi_dr)

    postadiabatic_dynamics = np.c_[t, r, phi, pr, pphi]


    return postadiabatic_dynamics


@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.initializedcheck(False)
@cython.profile(True)
@cython.linetrace(True)
cpdef compute_combined_dynamics(
    double omega0,
    Hamiltonian_C H,
    RadiationReactionForce RR,
    double chi_1,
    double chi_2,
    double m_1,
    double m_2,
    double tol=1e-12,
    EOBParams params=None,
    double step_back=50,
    str backend="ode",
    int PA_order=8,
    double r_stop=-1
):
    """
    Compute the full inspiral dynamics by combining PA + ODE
    integration. If the PA procedure fails (e.g. because the
    inspiral is too short), the code falls back to ODE
    integration.
    """
    try:
        postadiabatic_dynamics = compute_postadiabatic_dynamics(
            omega0,
            H,
            RR,
            chi_1,
            chi_2,
            m_1,
            m_2,
            tol=tol,
            params=params,
            order=PA_order
        )
        PA_success = True
        ode_y_init = postadiabatic_dynamics[-1, 1:]
    except ValueError as e:
        PA_success = False
        ode_y_init = None



    ode_dynamics_low, ode_dynamics_high = compute_dynamics(
        omega0,
        H,
        RR,
        chi_1,
        chi_2,
        m_1,
        m_2,
        rtol=1e-11,
        atol=1e-12,
        backend=backend,
        params=params,
        step_back=step_back,
        max_step=0.5,
        min_step=1.0e-9,
        y_init=ode_y_init,
        r_stop=r_stop
    )

    if PA_success is True:
        postadiabatic_dynamics = augment_dynamics(postadiabatic_dynamics,chi_1,chi_2,m_1,m_2,H)
        ode_dynamics_low[:, 0] += postadiabatic_dynamics[-1, 0]
        ode_dynamics_high[:, 0] += postadiabatic_dynamics[-1, 0]
        combined_dynamics = np.vstack(
            (postadiabatic_dynamics[:-1], ode_dynamics_low)
        )
    else:
        combined_dynamics = ode_dynamics_low

    return combined_dynamics, ode_dynamics_high


@cython.profile(True)
@cython.linetrace(True)
cpdef cumulative_integral(
    x: np.array,
    y: np.array,
    order: int = 7,
):
    """
    Compute a cumulative integral numerically using sampled data
    to a given order in accuracy. Assumes an equally spaced grid
    """
    cdef double h = x[1] - x[0]

    integral = np.zeros(x.size)

    if order == 3:
        for i in range(x.size - 1):
            if i == 0:
                z = np.sum(interpolated_integral_order_3[0] * y[:4])
            elif i == x.size - 2:
                z = np.sum(interpolated_integral_order_3[2] * y[-4:])
            else:
                z = np.sum(interpolated_integral_order_3[1] * y[i-1:i+3])

            integral[i+1] = integral[i] + z * h
    elif order == 5:
        for i in range(x.size - 1):
            if i == 0:
                z = np.sum(interpolated_integral_order_5[0] * y[:6])
            elif i == 1:
                z = np.sum(interpolated_integral_order_5[1] * y[:6])
            elif i == x.size - 3:
                z = np.sum(interpolated_integral_order_5[3] * y[-6:])
            elif i == x.size - 2:
                z = np.sum(interpolated_integral_order_5[4] * y[-6:])
            else:
                z = np.sum(interpolated_integral_order_5[2] * y[i-2:i+4])

            integral[i+1] = integral[i] + z * h
    elif order == 7:
        for i in range(x.size - 1):
            if i == 0:
                z = np.sum(interpolated_integral_order_7[0] * y[:8])
            elif i == 1:
                z = np.sum(interpolated_integral_order_7[1] * y[:8])
            elif i == 2:
                z = np.sum(interpolated_integral_order_7[2] * y[:8])
            elif i == x.size - 4:
                z = np.sum(interpolated_integral_order_7[4] * y[-8:])
            elif i == x.size - 3:
                z = np.sum(interpolated_integral_order_7[5] * y[-8:])
            elif i == x.size - 2:
                z = np.sum(interpolated_integral_order_7[6] * y[-8:])
            else:
                z = np.sum(interpolated_integral_order_7[3] * y[i-3:i+5])

            integral[i+1] = integral[i] + z * h

    return integral

@cython.profile(True)
@cython.linetrace(True)
def univariate_spline_integral(
    x: np.array,
    y: np.array,
) -> np.array:
    y_x_interp = InterpolatedUnivariateSpline(x[::-1], y[::-1])
    y_x_integral = y_x_interp.antiderivative()(x[::-1])[::-1]
    integral = y_x_integral - y_x_integral[0]

    return integral

@cython.profile(True)
@cython.linetrace(True)
def compute_adiabatic_parameter(
    dynamics,
    H,
    chi_1,
    chi_2,
    m_1,
    m_2,
    params,
):
    """
    Compute the adiabatic parameter \dot{\Omega}/2\Omega^{2}
    """
    omega = np.zeros(dynamics.shape[0])
    dr_dt = np.zeros(dynamics.shape[0])

    for i in range(dynamics.shape[0]):
        q = dynamics[i, 1:3]
        p = dynamics[i, 3:]

        dyn = H.dynamics(q, p, chi_1, chi_2, m_1, m_2)
        dH_dpr = dyn[2]


        csi = dyn[5]
        dr_dt[i] = dH_dpr * csi

        omega[i] = dyn[-1]

    domega_dr = fin_diff_derivative(dynamics[:, 1], omega)

    adiabatic_param = dr_dt * domega_dr / (2 * omega * omega)

    return adiabatic_param



from setuptools import setup,Extension
from Cython.Build import cythonize
from setuptools import find_packages
import numpy as np



extensions = [
    Extension("pyseobnr.eob.utils.containers", ["pyseobnr/eob/utils/containers.pyx"],
        include_dirs=[np.get_include()],
    ),
    Extension("pyseobnr.eob.waveform.waveform", ["pyseobnr/eob/waveform/waveform.pyx"],
              include_dirs=[np.get_include(),"pyseobnr/eob/utils"],
              extra_compile_args=['-O3'],
    ),
    Extension("pyseobnr.eob.hamiltonian.Ham_align_a6_apm_AP15_DP23_gaugeL_Tay_C", ["pyseobnr/eob/hamiltonian/Ham_align_a6_apm_AP15_DP23_gaugeL_Tay_C.pyx"],
                include_dirs=[np.get_include(),"pyseobnr/eob/utils"],
              extra_compile_args=['-O3'], define_macros=[('CYTHON_TRACE', '1')]
     ),
    Extension("pyseobnr.eob.hamiltonian.Hamiltonian_C", ["pyseobnr/eob/hamiltonian/Hamiltonian_C.pyx"],
              include_dirs=[np.get_include(),"pyseobnr/eob/utils"],
              extra_compile_args=['-O3'],
    ),
    Extension("pyseobnr.eob.dynamics.rhs_aligned", ["pyseobnr/eob/dynamics/rhs_aligned.pyx"],
              include_dirs=[np.get_include(),"pyseobnr/eob/utils"],
              extra_compile_args=['-O3'],
              define_macros=[('CYTHON_TRACE', '1')]
    ),
    Extension("pyseobnr.eob.dynamics.rhs_precessing", ["pyseobnr/eob/dynamics/rhs_precessing.pyx"],
              include_dirs=[np.get_include(),"pyseobnr/eob/utils"],
              extra_compile_args=['-O3']
    ),
    Extension("pyseobnr.eob.dynamics.postadiabatic_C", ["pyseobnr/eob/dynamics/postadiabatic_C.pyx"],
              include_dirs=[np.get_include(),"pyseobnr/eob/utils","pyseobnr/eob/hamiltonian"],
              extra_compile_args=['-O3']
    ),
    Extension("pyseobnr.eob.dynamics.postadiabatic_C_prec", ["pyseobnr/eob/dynamics/postadiabatic_C_prec.pyx"],
              include_dirs=[np.get_include(),"pyseobnr/eob/utils","pyseobnr/eob/hamiltonian"],
              extra_compile_args=['-O3']
    ),
    Extension("pyseobnr.eob.dynamics.postadiabatic_C_fast", ["pyseobnr/eob/dynamics/postadiabatic_C_fast.pyx"],
              include_dirs=[np.get_include(),"pyseobnr/eob/utils","pyseobnr/eob/hamiltonian"],
              extra_compile_args=['-O3']
              ),
    Extension("pyseobnr.eob.hamiltonian.Hamiltonian_v5PHM_C", ["pyseobnr/eob/hamiltonian/Hamiltonian_v5PHM_C.pyx"],
             include_dirs=[np.get_include(),"pyseobnr/eob/utils","pyseobnr/eob/hamiltonian"],
             extra_compile_args=['-O3']
    ),
    Extension("pyseobnr.eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD", ["pyseobnr/eob/hamiltonian/Ham_AvgS2precess_simple_cython_PA_AD.pyx"],
             include_dirs=[np.get_include(),"pyseobnr/eob/utils","pyseobnr/eob/hamiltonian"],
             extra_compile_args=['-O3'],
    ),
]



setup(
    packages=find_packages(),
    ext_modules=cythonize(extensions),
    zip_safe=False,
)



import numpy as np
from scipy.interpolate import CubicSpline

class VectorSpline:
    """This is just a cubic spline for every
    vector component separetly. Added to make
    it clear what is happening
    """
    def __init__(self, t:np.ndarray, v:np.ndarray):
        self.spline = CubicSpline(t, v)

    def __call__(self, t):
        return self.spline(t)
